# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import hashlib
from datetime import datetime
from datetime import timedelta
from datetime import timezone

from ..apprise_attachment import AppriseAttachment
from ..asset import AppriseAsset
from ..logger import logger
from ..exception import ApprisePluginException

try:
    import pgpy
    # Pretty Good Privacy (PGP) Support enabled
    PGP_SUPPORT = True

except ImportError:
    # Pretty Good Privacy (PGP) Support disabled
    PGP_SUPPORT = False


class ApprisePGPException(ApprisePluginException):
    """
    Thrown when there is an error with the Pretty Good Privacy Controller
    """
    def __init__(self, message, error_code=602):
        super().__init__(message, error_code=error_code)


class ApprisePGPController:
    """
    Pretty Good Privacy Controller Tool for the Apprise Library
    """

    # There is no reason a PGP Public Key should exceed 8K in size
    # If it is more than this, then it is not accepted
    max_pgp_public_key_size = 8000

    def __init__(self, path, pub_keyfile=None, email=None, asset=None,
                 **kwargs):
        """
        Path should be the directory keys can be written and read from such as
            <notifyobject>.store.path

            Optionally additionally specify a keyfile to explicitly open
        """

        # PGP hash
        self.__key_lookup = {}

        # Directory we can work with
        self.path = path

        # Our email
        self.email = email

        # Prepare our Asset Object
        self.asset = \
            asset if isinstance(asset, AppriseAsset) else AppriseAsset()

        if pub_keyfile:
            # Create ourselves an Attachment to work with; this grants us the
            # ability to pull this key from a remote site or anything else
            # supported by the Attachment object
            self._pub_keyfile = AppriseAttachment(asset=self.asset)

            # Add our definition to our pgp_key reference
            self._pub_keyfile.add(pub_keyfile)

            # Enforce maximum file size
            self._pub_keyfile[0].max_file_size = self.max_pgp_public_key_size

        else:
            self._pub_keyfile = None

    def keygen(self, email=None, name=None, force=False):
        """
        Generates a set of keys based on email configured.
        """

        try:
            # Create a new RSA key pair with 2048-bit strength
            key = pgpy.PGPKey.new(
                pgpy.constants.PubKeyAlgorithm.RSAEncryptOrSign, 2048)

        except NameError:
            # PGPy not installed
            logger.debug('PGPy not installed; keygen disabled')
            return False

        if self._pub_keyfile is not None or not self.path:
            logger.trace(
                'PGP keygen disabled, reason=%s',
                'keyfile-defined' if self._pub_keyfile is not None
                else 'no-write-path')
            return False

        if not name:
            name = self.asset.app_id

        if not email:
            email = self.email

        # Prepare our UID
        uid = pgpy.PGPUID.new(name, email=email)

        # Filenames
        file_prefix = email.split('@')[0].lower()

        pub_path = os.path.join(self.path, f'{file_prefix}-pub.asc')
        prv_path = os.path.join(self.path, f'{file_prefix}-prv.asc')

        if os.path.isfile(pub_path) and not force:
            logger.debug(
                'PGP generation skipped; Public Key already exists: %s',
                pub_path)
            return True

        # Persistent Storage Key
        lookup_key = hashlib.sha1(
            os.path.abspath(pub_path).encode('utf-8')).hexdigest()
        if lookup_key in self.__key_lookup:
            # Ensure our key no longer exists
            del self.__key_lookup[lookup_key]

        # Add the user ID to the key
        key.add_uid(uid, usage={
            pgpy.constants.KeyFlags.Sign,
            pgpy.constants.KeyFlags.EncryptCommunications},
            hashes=[pgpy.constants.HashAlgorithm.SHA256],
            ciphers=[pgpy.constants.SymmetricKeyAlgorithm.AES256],
            compression=[pgpy.constants.CompressionAlgorithm.ZLIB])

        try:
            # Write our keys to disk
            with open(pub_path, 'w') as f:
                f.write(str(key.pubkey))

        except OSError as e:
            logger.warning('Error writing PGP file %s', pub_path)
            logger.debug(f'I/O Exception: {e}')

            # Cleanup
            try:
                os.unlink(pub_path)
                logger.trace('Removed %s', pub_path)

            except OSError:
                pass

        try:
            with open(prv_path, 'w') as f:
                f.write(str(key))

        except OSError as e:
            logger.warning('Error writing PGP file %s', prv_path)
            logger.debug(f'I/O Exception: {e}')

            try:
                os.unlink(pub_path)
                logger.trace('Removed %s', pub_path)

            except OSError:
                pass

            try:
                os.unlink(prv_path)
                logger.trace('Removed %s', prv_path)

            except OSError:
                pass

            return False

        logger.info(
            'Wrote PGP Keys for %s/%s',
            os.path.dirname(pub_path),
            os.path.basename(pub_path))
        return True

    def public_keyfile(self, *emails):
        """
        Returns the first match of a useable public key based emails provided
        """

        if not PGP_SUPPORT:
            msg = 'PGP Support unavailable; install PGPy library'
            logger.warning(msg)
            raise ApprisePGPException(msg)

        if self._pub_keyfile is not None:
            # If our code reaches here, then we fetch our public key
            pgp_key = self._pub_keyfile[0]
            if not pgp_key:
                # We could not access the attachment
                logger.error(
                    'Could not access PGP Public Key {}.'.format(
                        pgp_key.url(privacy=True)))
                return False

            return pgp_key.path

        elif not self.path:
            # No path
            return None

        fnames = [
            'pgp-public.asc',
            'pgp-pub.asc',
            'public.asc',
            'pub.asc',
        ]

        if self.email:
            # Include our email in the list
            emails = [self.email] + [*emails]

        for email in emails:
            _entry = email.split('@')[0].lower()
            fnames.insert(0, f'{_entry}-pub.asc')

            # Lowercase email (Highest Priority)
            _entry = email.lower()
            fnames.insert(0, f'{_entry}-pub.asc')

        return next(
            (os.path.join(self.path, fname)
             for fname in fnames
             if os.path.isfile(os.path.join(self.path, fname))),
            None)

    def public_key(self, *emails, autogen=None):
        """
        Opens a spcified pgp public file and returns the key from it which
        is used to encrypt the message
        """
        path = self.public_keyfile(*emails)
        if not path:
            if (autogen if autogen is not None else self.asset.pgp_autogen) \
                    and self.keygen(*emails):
                path = self.public_keyfile(*emails)
                if path:
                    # We should get a hit now
                    return self.public_key(*emails)

            logger.warning('No PGP Public Key could be loaded')
            return None

        # Persistent Storage Key
        key = hashlib.sha1(
            os.path.abspath(path).encode('utf-8')).hexdigest()
        if key in self.__key_lookup:
            # Take an early exit
            return self.__key_lookup[key]['public_key']

        try:
            with open(path, 'r') as key_file:
                public_key, _ = pgpy.PGPKey.from_blob(key_file.read())

        except NameError:
            # PGPy not installed
            logger.debug(
                'PGPy not installed; skipping PGP support: %s', path)
            return None

        except FileNotFoundError:
            # Generate keys
            logger.debug('PGP Public Key file not found: %s', path)
            return None

        except OSError as e:
            logger.warning('Error accessing PGP Public Key file %s', path)
            logger.debug(f'I/O Exception: {e}')
            return None

        self.__key_lookup[key] = {
            'public_key': public_key,
            'expires':
            datetime.now(timezone.utc) + timedelta(seconds=86400)
        }
        return public_key

    # Encrypt message using the recipient's public key
    def encrypt(self, message, *emails):
        """
        If provided a path to a pgp-key, content is encrypted
        """

        # Acquire our key
        public_key = self.public_key(*emails)
        if not public_key:
            # Encryption not possible
            return False

        try:
            message_object = pgpy.PGPMessage.new(message)
            encrypted_message = public_key.encrypt(message_object)
            return str(encrypted_message)

        except pgpy.errors.PGPError:
            # Encryption not Possible
            logger.debug(
                'PGP Public Key Corruption; encryption not possible')

        except NameError:
            # PGPy not installed
            logger.debug('PGPy not installed; Skipping PGP encryption')

        return None

    def prune(self):
        """
        Prunes old entries from the public_key index
        """
        self.__key_lookup = {
            key: value for key, value in self.__key_lookup.items()
            if value['expires'] > datetime.now(timezone.utc)}

    @property
    def pub_keyfile(self):
        """
        Returns the Public Keyfile Path if set otherwise it returns None
        This property returns False if a keyfile was provided, but was invalid
        """
        return None if self._pub_keyfile is None else (
            False if not self._pub_keyfile[0] else self._pub_keyfile[0].path)



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re
import requests
from json import dumps, loads
from itertools import chain

from .base import NotifyBase
from ..common import NotifyType
from .. import exception
from ..utils.parse import (
    validate_regex, is_phone_no, parse_phone_no, parse_bool)
from ..url import PrivacyMode
from ..locale import gettext_lazy as _


GROUP_REGEX = re.compile(
    r'^\s*(\#|\%35)(?P<group>[a-z0-9_-]+)', re.I)

CONTACT_REGEX = re.compile(
    r'^\s*(\@|\%40)?(?P<contact>[a-z0-9_-]+)', re.I)


# Priorities
class SMSEaglePriority:
    NORMAL = 0
    HIGH = 1


SMSEAGLE_PRIORITIES = (
    SMSEaglePriority.NORMAL,
    SMSEaglePriority.HIGH,
)

SMSEAGLE_PRIORITY_MAP = {
    # short for 'normal'
    'normal': SMSEaglePriority.NORMAL,
    # short for 'high'
    '+': SMSEaglePriority.HIGH,
    'high': SMSEaglePriority.HIGH,
}


class SMSEagleCategory:
    """
    We define the different category types that we can notify via SMS Eagle
    """
    PHONE = 'phone'
    GROUP = 'group'
    CONTACT = 'contact'


SMSEAGLE_CATEGORIES = (
    SMSEagleCategory.PHONE,
    SMSEagleCategory.GROUP,
    SMSEagleCategory.CONTACT,
)


class NotifySMSEagle(NotifyBase):
    """
    A wrapper for SMSEagle Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'SMS Eagle'

    # The services URL
    service_url = 'https://smseagle.eu'

    # The default protocol
    protocol = 'smseagle'

    # The default protocol
    secure_protocol = 'smseagles'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_smseagle'

    # The path we send our notification to
    notify_path = '/jsonrpc/sms'

    # Support attachments
    attachment_support = True

    # The maxumum length of the text message
    # The actual limit is 160 but SMSEagle looks after the handling
    # of large messages in it's upstream service
    body_maxlen = 1200

    # The maximum targets to include when doing batch transfers
    default_batch_size = 10

    # We don't support titles for SMSEagle notifications
    title_maxlen = 0

    # Define object templates
    templates = (
        '{schema}://{token}@{host}/{targets}',
        '{schema}://{token}@{host}:{port}/{targets}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'host': {
            'name': _('Hostname'),
            'type': 'string',
            'required': True,
        },
        'port': {
            'name': _('Port'),
            'type': 'int',
            'min': 1,
            'max': 65535,
        },
        'token': {
            'name': _('Access Token'),
            'type': 'string',
            'required': True,
        },
        'target_phone': {
            'name': _('Target Phone No'),
            'type': 'string',
            'prefix': '+',
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
            'map_to': 'targets',
        },
        'target_group': {
            'name': _('Target Group ID'),
            'type': 'string',
            'prefix': '#',
            'regex': (r'^[a-z0-9_-]+$', 'i'),
            'map_to': 'targets',
        },
        'target_contact': {
            'name': _('Target Contact'),
            'type': 'string',
            'prefix': '@',
            'regex': (r'^[a-z0-9_-]+$', 'i'),
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
            'required': True,
        }
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'to': {
            'alias_of': 'targets',
        },
        'token': {
            'alias_of': 'token',
        },
        'batch': {
            'name': _('Batch Mode'),
            'type': 'bool',
            'default': False,
        },
        'status': {
            'name': _('Show Status'),
            'type': 'bool',
            'default': False,
        },
        'test': {
            'name': _('Test Only'),
            'type': 'bool',
            'default': False,
        },
        'flash': {
            'name': _('Flash'),
            'type': 'bool',
            'default': False,
        },
        'priority': {
            'name': _('Priority'),
            'type': 'choice:int',
            'values': SMSEAGLE_PRIORITIES,
            'default': SMSEaglePriority.NORMAL,
        },
    })

    def __init__(self, token=None, targets=None, priority=None, batch=False,
                 status=False, flash=False, test=False, **kwargs):
        """
        Initialize SMSEagle Object
        """
        super().__init__(**kwargs)

        # Prepare Flash Mode Flag
        self.flash = flash

        # Prepare Test Mode Flag
        self.test = test

        # Prepare Batch Mode Flag
        self.batch = batch

        # Set Status type
        self.status = status

        # Parse our targets
        self.target_phones = list()
        self.target_groups = list()
        self.target_contacts = list()

        # Used for URL generation afterwards only
        self.invalid_targets = list()

        # We always use a token if provided
        self.token = validate_regex(self.user if not token else token)
        if not self.token:
            msg = \
                'An invalid SMSEagle Access Token ({}) was specified.'.format(
                    self.user if not token else token)
            self.logger.warning(msg)
            raise TypeError(msg)

        #
        # Priority
        #
        try:
            # Acquire our priority if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.priority = int(priority)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.priority = self.template_args['priority']['default']

        except ValueError:
            # Input is a string; attempt to get the lookup from our
            # priority mapping
            priority = priority.lower().strip()

            # This little bit of black magic allows us to match against
            # low, lo, l (for low);
            # normal, norma, norm, nor, no, n (for normal)
            # ... etc
            result = next((key for key in SMSEAGLE_PRIORITY_MAP.keys()
                          if key.startswith(priority)), None) \
                if priority else None

            # Now test to see if we got a match
            if not result:
                msg = 'An invalid SMSEagle priority ' \
                      '({}) was specified.'.format(priority)
                self.logger.warning(msg)
                raise TypeError(msg)

            # store our successfully looked up priority
            self.priority = SMSEAGLE_PRIORITY_MAP[result]

        if self.priority is not None and \
                self.priority not in SMSEAGLE_PRIORITY_MAP.values():
            msg = 'An invalid SMSEagle priority ' \
                  '({}) was specified.'.format(priority)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate our targerts
        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            # Allow 9 digit numbers (without country code)
            result = is_phone_no(target, min_len=9)
            if result:
                # store valid phone number
                self.target_phones.append(
                    '{}{}'.format(
                        '' if target[0] != '+' else '+', result['full']))
                continue

            result = GROUP_REGEX.match(target)
            if result:
                # Just store group information
                self.target_groups.append(result.group('group'))
                continue

            result = CONTACT_REGEX.match(target)
            if result:
                # Just store contact information
                self.target_contacts.append(result.group('contact'))
                continue

            self.logger.warning(
                'Dropped invalid phone/group/contact '
                '({}) specified.'.format(target),
            )
            self.invalid_targets.append(target)
            continue

        return

    def send(self, body, title='', notify_type=NotifyType.INFO, attach=None,
             **kwargs):
        """
        Perform SMSEagle Notification
        """

        if not self.target_groups and not self.target_phones \
                and not self.target_contacts:
            # There were no services to notify
            self.logger.warning(
                'There were no SMSEagle targets to notify.')
            return False

        # error tracking (used for function return)
        has_error = False

        attachments = []
        if attach and self.attachment_support:
            for attachment in attach:
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        'Could not access SMSEagle attachment {}.'.format(
                            attachment.url(privacy=True)))
                    return False

                if not re.match(r'^image/.*', attachment.mimetype, re.I):
                    # Only support images at this time
                    self.logger.warning(
                        'Ignoring unsupported SMSEagle attachment {}.'.format(
                            attachment.url(privacy=True)))
                    continue

                try:
                    # Prepare our Attachment in Base64
                    attachments.append({
                        'content_type': attachment.mimetype,
                        'content': attachment.base64(),
                    })

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        'Could not access SMSEagle attachment {}.'.format(
                            attachment.url(privacy=True)))
                    return False

                self.logger.debug(
                    'Appending SMSEagle attachment {}'.format(
                        attachment.url(privacy=True)))

        # Prepare our headers
        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/json',
        }

        # Prepare our payload
        params_template = {
            # Our Access Token
            'access_token': self.token,

            # The message to send (populated below)
            "message": None,

            # 0 = normal priority, 1 = high priority
            "highpriority": self.priority,

            # Support unicode characters
            "unicode": 1,

            # sms or mms (if attachment)
            "message_type": 'sms',

            # Response Types:
            #  simple: format response as simple object with one result field
            #  extended: format response as extended JSON object
            "responsetype": 'extended',

            # SMS will be sent as flash message (1 = yes, 0 = no)
            "flash": 1 if self.flash else 0,

            # Message Simulation
            "test": 1 if self.test else 0,
        }

        # Set our schema
        schema = 'https' if self.secure else 'http'

        # Construct our URL
        notify_url = '%s://%s' % (schema, self.host)
        if isinstance(self.port, int):
            notify_url += ':%d' % self.port
        notify_url += self.notify_path

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        notify_by = {
            SMSEagleCategory.PHONE: {
                "method": "sms.send_sms",
                'target': 'to',
            },
            SMSEagleCategory.GROUP: {
                "method": "sms.send_togroup",
                'target': 'groupname',
            },
            SMSEagleCategory.CONTACT: {
                "method": "sms.send_tocontact",
                'target': 'contactname',
            },
        }

        # categories separated into a tuple since notify_by.keys()
        # returns an unpredicable list in Python 2.7 which causes
        # tests to fail every so often
        for category in SMSEAGLE_CATEGORIES:
            # Create a copy of our template
            payload = {
                'method': notify_by[category]['method'],
                'params': {
                    notify_by[category]['target']: None,
                },
            }

            # Apply Template
            payload['params'].update(params_template)

            # Set our Message
            payload["params"]["message"] = "{}{}".format(
                '' if not self.status else '{} '.format(
                    self.asset.ascii(notify_type)), body)

            if attachments:
                # Store our attachments
                payload['params']['message_type'] = 'mms'
                payload['params']['attachments'] = attachments

            targets = getattr(self, 'target_{}s'.format(category))
            for index in range(0, len(targets), batch_size):
                # Prepare our recipients
                payload['params'][notify_by[category]['target']] = \
                    ','.join(targets[index:index + batch_size])

                self.logger.debug('SMSEagle POST URL: %s (cert_verify=%r)' % (
                    notify_url, self.verify_certificate,
                ))
                self.logger.debug('SMSEagle Payload: %s' % str(payload))

                # Always call throttle before any remote server i/o is made
                self.throttle()
                try:
                    r = requests.post(
                        notify_url,
                        data=dumps(payload),
                        headers=headers,
                        verify=self.verify_certificate,
                        timeout=self.request_timeout,
                    )

                    try:
                        content = loads(r.content)

                        # Store our status
                        status_str = str(content['result'])

                    except (AttributeError, TypeError, ValueError, KeyError):
                        # ValueError = r.content is Unparsable
                        # TypeError = r.content is None
                        # AttributeError = r is None
                        # KeyError = 'result' is not found in result
                        content = {}

                    # The result set can be a list such as:
                    #   b'{"result":[{"message_id":4753,"status":"ok"}]}'
                    #
                    # It can also just be as a dictionary:
                    #   b'{"result":{"message_id":4753,"status":"ok"}}'
                    #
                    # The below code handles both cases only only fails if a
                    # non-ok value was returned

                    if r.status_code not in (
                            requests.codes.ok, requests.codes.created) or \
                            not isinstance(content.get('result'),
                                           (dict, list)) or \
                            (isinstance(content.get('result'), dict) and
                             content['result'].get('status') != 'ok') or \
                            (isinstance(content.get('result'), list) and
                             next((True for entry in content.get('result')
                                   if isinstance(entry, dict) and
                                   entry.get('status') != 'ok'), False
                                  )  # pragma: no cover
                             ):

                        # We had a problem
                        status_str = content.get('result') \
                            if content.get('result') else \
                            NotifySMSEagle.http_response_code_lookup(
                            r.status_code)

                        self.logger.warning(
                            'Failed to send {} {} SMSEagle {} notification: '
                            '{}{}error={}.'.format(
                                len(targets[index:index + batch_size]),
                                'to {}'.format(targets[index])
                                if batch_size == 1 else '(s)',
                                category,
                                status_str,
                                ', ' if status_str else '',
                                r.status_code))

                        self.logger.debug(
                            'Response {} Details:\r\n{}'.format(
                                category.upper(), r.content))

                        # Mark our failure
                        has_error = True
                        continue

                    else:
                        self.logger.info(
                            'Sent {} SMSEagle {} notification{}.'
                            .format(
                                len(targets[index:index + batch_size]),
                                category,
                                ' to {}'.format(targets[index])
                                if batch_size == 1 else '(s)',
                            ))

                except requests.RequestException as e:
                    self.logger.warning(
                        'A Connection error occured sending {} SMSEagle '
                        '{} notification(s).'.format(
                            len(targets[index:index + batch_size]), category))
                    self.logger.debug('Socket Exception: %s' % str(e))

                    # Mark our failure
                    has_error = True
                    continue

        return not has_error

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.token, self.host, self.port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            'batch': 'yes' if self.batch else 'no',
            'status': 'yes' if self.status else 'no',
            'flash': 'yes' if self.flash else 'no',
            'test': 'yes' if self.test else 'no',
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        default_priority = self.template_args['priority']['default']
        if self.priority is not None:
            # Store our priority; but only if it was specified
            params['priority'] = \
                next((key for key, value in SMSEAGLE_PRIORITY_MAP.items()
                      if value == self.priority),
                     default_priority)  # pragma: no cover

        # Default port handling
        default_port = 443 if self.secure else 80

        return '{schema}://{token}@{hostname}{port}/{targets}?{params}'.format(
            schema=self.secure_protocol if self.secure else self.protocol,
            token=self.pprint(
                self.token, privacy, mode=PrivacyMode.Secret, safe=''),
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port='' if self.port is None or self.port == default_port
                 else ':{}'.format(self.port),
            targets='/'.join(
                [NotifySMSEagle.quote(x, safe='#@') for x in chain(
                    # Pass phones directly as is
                    self.target_phones,
                    # Contacts
                    ['@{}'.format(x) for x in self.target_contacts],
                    # Groups
                    ['#{}'.format(x) for x in self.target_groups],
                    # Pass along the same invalid entries as were provided
                    self.invalid_targets,
                )]),
            params=NotifySMSEagle.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        #
        # Factor batch into calculation
        #
        batch_size = 1 if not self.batch else self.default_batch_size
        if batch_size > 1:
            # Batches can only be sent by group (you can't combine groups into
            # a single batch)
            total_targets = 0
            for c in SMSEAGLE_CATEGORIES:
                targets = len(getattr(self, f'target_{c}s'))
                total_targets += int(targets / batch_size) + \
                    (1 if targets % batch_size else 0)
            return total_targets

        # Normal batch count; just count the targets
        return len(self.target_phones) + len(self.target_contacts) + \
            len(self.target_groups)

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results['targets'] = \
            NotifySMSEagle.split_path(results['fullpath'])

        if 'token' in results['qsd'] and len(results['qsd']['token']):
            results['token'] = NotifySMSEagle.unquote(results['qsd']['token'])

        elif not results['password'] and results['user']:
            results['token'] = NotifySMSEagle.unquote(results['user'])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'] += \
                NotifySMSEagle.parse_phone_no(results['qsd']['to'])

        # Get Batch Mode Flag
        results['batch'] = \
            parse_bool(results['qsd'].get('batch', False))

        # Get Flash Mode Flag
        results['flash'] = \
            parse_bool(results['qsd'].get('flash', False))

        # Get Test Mode Flag
        results['test'] = \
            parse_bool(results['qsd'].get('test', False))

        # Get status switch
        results['status'] = \
            parse_bool(results['qsd'].get('status', False))

        # Get priority
        if 'priority' in results['qsd'] and len(results['qsd']['priority']):
            results['priority'] = \
                NotifySMSEagle.unquote(results['qsd']['priority'])

        return results



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this plugin, you need to create a MS Teams Azure Webhook Workflow:
#  https://support.microsoft.com/en-us/office/browse-and-add-workflows-\
#       in-microsoft-teams-4998095c-8b72-4b0e-984c-f2ad39e6ba9a

# Your webhook will look somthing like this:
# https://prod-161.westeurope.logic.azure.com:443/\
#       workflows/643e69f83c8944438d68119179a10a64/triggers/manual/\
#       paths/invoke?api-version=2016-06-01&sp=%2Ftriggers%2Fmanual%2Frun&\
#       sv=1.0&sig=KODuebWbDGYFr0z0eu-6Rj8aUKz7108W3wrNJZxFE5A
#
# Yes... The URL is that big... But it looks like this (greatly simplified):
# https://HOST:PORT/workflows/ABCD/triggers/manual/path/...sig=DEFG
#          ^    ^                ^                              ^
#          |    |                |                              |
#  These are important <---------^------------------------------^
#
#
# Apprise can support this webhook as is (directly passed into it)
# Alternatively it can be shortend to:

# These 3 tokens need to be placed in the URL after the Team
#   workflows://HOST:PORT/ABCD/DEFG/
#

import re
import requests
import json
from json.decoder import JSONDecodeError

from .base import NotifyBase
from ..common import NotifyImageSize
from ..common import NotifyType
from ..common import NotifyFormat
from ..utils.parse import parse_bool, validate_regex
from ..utils.templates import apply_template, TemplateType
from ..apprise_attachment import AppriseAttachment
from ..locale import gettext_lazy as _


class NotifyWorkflows(NotifyBase):
    """
    A wrapper for Microsoft Workflows (MS Teams) Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'Power Automate / Workflows (for MSTeams)'

    # The services URL
    service_url = 'https://www.microsoft.com/power-platform/' \
        'products/power-automate'

    # The default secure protocol
    secure_protocol = ('workflow', 'workflows')

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_workflows'

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_32

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 1000

    # Default Notification Format
    notify_format = NotifyFormat.MARKDOWN

    # There is no reason we should exceed 35KB when reading in a JSON file.
    # If it is more than this, then it is not accepted
    max_workflows_template_size = 35000

    # Adaptive Card Version
    adaptive_card_version = '1.4'

    # Define object templates
    templates = (
        '{schema}://{host}/{workflow}/{signature}',
        '{schema}://{host}:{port}/{workflow}/{signature}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'host': {
            'name': _('Hostname'),
            'type': 'string',
            'required': True,
        },
        'port': {
            'name': _('Port'),
            'type': 'int',
            'min': 1,
            'max': 65535,
        },
        # workflow identifier
        'workflow': {
            'name': _('Workflow ID'),
            'type': 'string',
            'private': True,
            'required': True,
            'regex': (r'^[A-Z0-9_-]+$', 'i'),
        },
        # Signature
        'signature': {
            'name': _('Signature'),
            'type': 'string',
            'private': True,
            'required': True,
            'regex': (r'^[a-z0-9_-]+$', 'i'),
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'id': {
            'alias_of': 'workflow',
        },
        'image': {
            'name': _('Include Image'),
            'type': 'bool',
            'default': True,
            'map_to': 'include_image',
        },
        'wrap': {
            'name': _('Wrap Text'),
            'type': 'bool',
            'default': True,
            'map_to': 'wrap',
        },
        'template': {
            'name': _('Template Path'),
            'type': 'string',
            'private': True,
        },
        # Below variable shortforms are taken from the Workflows webhook
        # for consistency
        'sig': {
            'alias_of': 'signature',
        },
        'ver': {
            'name': _('API Version'),
            'type': 'string',
            'default': '2016-06-01',
            'map_to': 'version',
        },
        'api-version': {
            'alias_of': 'ver'
        },
    })

    # Define our token control
    template_kwargs = {
        'tokens': {
            'name': _('Template Tokens'),
            'prefix': ':',
        },
    }

    def __init__(self, workflow, signature, include_image=None,
                 version=None, template=None, tokens=None, wrap=None,
                 **kwargs):
        """
        Initialize Microsoft Workflows Object

        """
        super().__init__(**kwargs)

        self.workflow = validate_regex(
            workflow, *self.template_tokens['workflow']['regex'])
        if not self.workflow:
            msg = 'An invalid Workflows ID ' \
                  '({}) was specified.'.format(workflow)
            self.logger.warning(msg)
            raise TypeError(msg)

        self.signature = validate_regex(
            signature, *self.template_tokens['signature']['regex'])
        if not self.signature:
            msg = 'An invalid Signature ' \
                  '({}) was specified.'.format(signature)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Place a thumbnail image inline with the message body
        self.include_image = True if (
            include_image if include_image is not None else
            self.template_args['image']['default']) else False

        # Wrap Text
        self.wrap = True if (
            wrap if wrap is not None else
            self.template_args['wrap']['default']) else False

        # Our template object is just an AppriseAttachment object
        self.template = AppriseAttachment(asset=self.asset)
        if template:
            # Add our definition to our template
            self.template.add(template)
            # Enforce maximum file size
            self.template[0].max_file_size = self.max_workflows_template_size

        # Prepare Version
        self.api_version = version if version is not None \
            else self.template_args['ver']['default']

        # Template functionality
        self.tokens = {}
        if isinstance(tokens, dict):
            self.tokens.update(tokens)

        elif tokens:
            msg = 'The specified Workflows Template Tokens ' \
                  '({}) are not identified as a dictionary.'.format(tokens)
            self.logger.warning(msg)
            raise TypeError(msg)

        # else:  NoneType - this is okay
        return

    def gen_payload(self, body, title='', notify_type=NotifyType.INFO,
                    **kwargs):
        """
        This function generates our payload whether it be the generic one
        Apprise generates by default, or one provided by a specified
        external template.
        """

        # Acquire our to-be footer icon if configured to do so
        image_url = None if not self.include_image \
            else self.image_url(notify_type)

        body_content = []
        if image_url:
            body_content.append({
                "type": "Image",
                "url": image_url,
                "height": "32px",
                "altText": notify_type,
            })

        if title:
            body_content.append({
                "type": "TextBlock",
                "text": f'{title}',
                "style": "heading",
                "weight": "Bolder",
                "size": "Large",
                "id": "title",
            })

        body_content.append({
            "type": "TextBlock",
            "text": body,
            "style": "default",
            "wrap": self.wrap,
            "id": "body",
        })

        if not self.template:
            # By default we use a generic working payload if there was
            # no template specified
            schema = "http://adaptivecards.io/schemas/adaptive-card.json"
            payload = {
                "type": "message",
                "attachments": [
                    {
                        "contentType":
                        "application/vnd.microsoft.card.adaptive",
                        "contentUrl": None,
                        "content": {
                            "$schema": schema,
                            "type": "AdaptiveCard",
                            "version": self.adaptive_card_version,
                            "body": body_content,
                            # Additionally
                            "msteams": {"width": "full"},
                        }
                    }
                ]
            }

            return payload

        # If our code reaches here, then we generate ourselves the payload
        template = self.template[0]
        if not template:
            # We could not access the attachment
            self.logger.error(
                'Could not access Workflow template {}.'.format(
                    template.url(privacy=True)))
            return False

        # Take a copy of our token dictionary
        tokens = self.tokens.copy()

        # Apply some defaults template values
        tokens['app_body'] = body
        tokens['app_title'] = title
        tokens['app_type'] = notify_type
        tokens['app_id'] = self.app_id
        tokens['app_desc'] = self.app_desc
        tokens['app_color'] = self.color(notify_type)
        tokens['app_image_url'] = image_url
        tokens['app_url'] = self.app_url

        # Enforce Application mode
        tokens['app_mode'] = TemplateType.JSON

        try:
            with open(template.path, 'r') as fp:
                content = json.loads(apply_template(fp.read(), **tokens))

        except (OSError, IOError):
            self.logger.error(
                'MSTeam template {} could not be read.'.format(
                    template.url(privacy=True)))
            return None

        except JSONDecodeError as e:
            self.logger.error(
                'MSTeam template {} contains invalid JSON.'.format(
                    template.url(privacy=True)))
            self.logger.debug('JSONDecodeError: {}'.format(e))
            return None

        return content

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform Microsoft Teams Notification
        """

        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/json',
        }

        params = {
            'api-version': self.api_version,
            'sp': '/triggers/manual/run',
            'sv': '1.0',
            'sig': self.signature,
        }

        notify_url = 'https://{host}{port}/workflows/{workflow}/' \
            'triggers/manual/paths/invoke'.format(
                host=self.host,
                port='' if not self.port else f':{self.port}',
                workflow=self.workflow)

        # Generate our payload if it's possible
        payload = self.gen_payload(
            body=body, title=title, notify_type=notify_type, **kwargs)
        if not payload:
            # No need to present a reason; that will come from the
            # gen_payload() function itself
            return False

        self.logger.debug('Workflows POST URL: %s (cert_verify=%r)' % (
            notify_url, self.verify_certificate,
        ))
        self.logger.debug('Workflows Payload: %s' % str(payload))

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                notify_url,
                params=params,
                data=json.dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                    requests.codes.ok, requests.codes.accepted):
                # We had a problem
                status_str = \
                    NotifyWorkflows.http_response_code_lookup(r.status_code)

                self.logger.warning(
                    'Failed to send Workflows notification: '
                    '{}{}error={}.'.format(
                        status_str,
                        ', ' if status_str else '',
                        r.status_code))

                self.logger.debug(
                    'Response Details:\r\n{}'.format(r.content))

                # We failed
                return False

            else:
                self.logger.info('Sent Workflows notification.')

        except requests.RequestException as e:
            self.logger.warning(
                'A Connection error occurred sending Workflows notification.')
            self.logger.debug('Socket Exception: %s' % str(e))

            # We failed
            return False

        return True

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol[0], self.host, self.port, self.workflow,
            self.signature,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            'image': 'yes' if self.include_image else 'no',
            'wrap': 'yes' if self.wrap else 'no',
        }

        if self.template:
            params['template'] = NotifyWorkflows.quote(
                self.template[0].url(), safe='')

        # Store our version if it differs from default
        if self.api_version != self.template_args['ver']['default']:
            params['ver'] = self.api_version

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))
        # Store any template entries if specified
        params.update({':{}'.format(k): v for k, v in self.tokens.items()})

        return '{schema}://{host}{port}/{workflow}/{signature}/' \
            '?{params}'.format(
                schema=self.secure_protocol[0],
                host=self.host,
                port='' if not self.port else f':{self.port}',
                workflow=self.pprint(self.workflow, privacy, safe=''),
                signature=self.pprint(self.signature, privacy, safe=''),
                params=NotifyWorkflows.urlencode(params),
            )

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # store values if provided
        entries = NotifyWorkflows.split_path(results['fullpath'])

        # Display image?
        results['include_image'] = parse_bool(results['qsd'].get(
            'image', NotifyWorkflows.template_args['image']['default']))

        # Wrap Text?
        results['wrap'] = parse_bool(results['qsd'].get(
            'wrap', NotifyWorkflows.template_args['wrap']['default']))

        # Template Handling
        if 'template' in results['qsd'] and results['qsd']['template']:
            results['template'] = \
                NotifyWorkflows.unquote(results['qsd']['template'])

        if 'workflow' in results['qsd'] and results['qsd']['workflow']:
            results['workflow'] = \
                NotifyWorkflows.unquote(results['qsd']['workflow'])

        elif 'id' in results['qsd'] and results['qsd']['id']:
            results['workflow'] = \
                NotifyWorkflows.unquote(results['qsd']['id'])

        else:
            results['workflow'] = None if not entries \
                else NotifyWorkflows.unquote(entries.pop(0))

        # Signature
        if 'signature' in results['qsd'] and results['qsd']['signature']:
            results['signature'] = \
                NotifyWorkflows.unquote(results['qsd']['signature'])

        elif 'sig' in results['qsd'] and results['qsd']['sig']:
            results['signature'] = \
                NotifyWorkflows.unquote(results['qsd']['sig'])

        else:
            # Read information from path
            results['signature'] = None if not entries \
                else NotifyWorkflows.unquote(entries.pop(0))

        # Version
        if 'api-version' in results['qsd'] and results['qsd']['api-version']:
            results['version'] = \
                NotifyWorkflows.unquote(results['qsd']['api-version'])

        elif 'ver' in results['qsd'] and results['qsd']['ver']:
            results['version'] = \
                NotifyWorkflows.unquote(results['qsd']['ver'])

        # Store our tokens
        results['tokens'] = results['qsd:']

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support parsing the webhook straight out of workflows
            https://HOST:443/workflows/WORKFLOWID/triggers/manual/paths/invoke
        """

        # Match our workflows webhook URL and re-assemble
        result = re.match(
            r'^https?://(?P<host>[A-Z0-9_.-]+)'
            r'(?P<port>:[1-9][0-9]{0,5})?'
            r'/workflows/'
            r'(?P<workflow>[A-Z0-9_-]+)'
            r'/triggers/manual/paths/invoke/?'
            r'(?P<params>\?.+)$', url, re.I)

        if result:
            # Construct our URL
            return NotifyWorkflows.parse_url(
                '{schema}://{host}{port}/{workflow}'
                '/{params}'.format(
                    schema=NotifyWorkflows.secure_protocol[0],
                    host=result.group('host'),
                    port='' if not result.group('port')
                    else result.group('port'),
                    workflow=result.group('workflow'),
                    params=result.group('params')))
        return None



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Notifico allows you to relay notifications into IRC channels.
#
# 1. visit https://n.tkte.ch and sign up for an account
# 2. create a project; either manually or sync with github
# 3. from within the project, you can create a message hook
#
# the URL will look something like this:
#       https://n.tkte.ch/h/2144/uJmKaBW9WFk42miB146ci3Kj
#                            ^                ^
#                            |                |
#                         project id       message hook
#
# This plugin also supports taking the URL (as identified above) directly
# as well.

import re
import requests

from .base import NotifyBase
from ..common import NotifyType
from ..utils.parse import parse_bool, validate_regex
from ..locale import gettext_lazy as _


class NotificoFormat:
    # Resets all formatting
    Reset = '\x0F'

    # Formatting
    Bold = '\x02'
    Italic = '\x1D'
    Underline = '\x1F'
    BGSwap = '\x16'


class NotificoColor:
    # Resets Color
    Reset = '\x03'

    # Colors
    White = '\x0300'
    Black = '\x0301'
    Blue = '\x0302'
    Green = '\x0303'
    Red = '\x0304'
    Brown = '\x0305'
    Purple = '\x0306'
    Orange = '\x0307'
    Yellow = '\x0308',
    LightGreen = '\x0309'
    Teal = '\x0310'
    LightCyan = '\x0311'
    LightBlue = '\x0312'
    Violet = '\x0313'
    Grey = '\x0314'
    LightGrey = '\x0315'


class NotifyNotifico(NotifyBase):
    """
    A wrapper for Notifico Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'Notifico'

    # The services URL
    service_url = 'https://n.tkte.ch'

    # The default protocol
    protocol = 'notifico'

    # The default secure protocol
    secure_protocol = 'notifico'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_notifico'

    # Plain Text Notification URL
    notify_url = 'https://n.tkte.ch/h/{proj}/{hook}'

    # The title is not used
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 512

    # Define object templates
    templates = (
        '{schema}://{project_id}/{msghook}',
    )

    # Define our template arguments
    template_tokens = dict(NotifyBase.template_tokens, **{
        # The Project ID is found as the first part of the URL
        #  /1234/........................
        'project_id': {
            'name': _('Project ID'),
            'type': 'string',
            'required': True,
            'private': True,
            'regex': (r'^[0-9]+$', ''),
        },
        # The Message Hook follows the Project ID
        #  /..../AbCdEfGhIjKlMnOpQrStUvWX
        'msghook': {
            'name': _('Message Hook'),
            'type': 'string',
            'required': True,
            'private': True,
            'regex': (r'^[a-z0-9]+$', 'i'),
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        # You can optionally pass IRC colors into
        'color': {
            'name': _('IRC Colors'),
            'type': 'bool',
            'default': True,
        },

        # You can optionally pass IRC color into
        'prefix': {
            'name': _('Prefix'),
            'type': 'bool',
            'default': True,
        },
    })

    def __init__(self, project_id, msghook, color=True, prefix=True,
                 **kwargs):
        """
        Initialize Notifico Object
        """
        super().__init__(**kwargs)

        # Assign our message hook
        self.project_id = validate_regex(
            project_id, *self.template_tokens['project_id']['regex'])
        if not self.project_id:
            msg = 'An invalid Notifico Project ID ' \
                  '({}) was specified.'.format(project_id)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Assign our message hook
        self.msghook = validate_regex(
            msghook, *self.template_tokens['msghook']['regex'])
        if not self.msghook:
            msg = 'An invalid Notifico Message Token ' \
                  '({}) was specified.'.format(msghook)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prefix messages with a [?] where ? identifies the message type
        # such as if it's an error, warning, info, or success
        self.prefix = prefix

        # Send colors
        self.color = color

        # Prepare our notification URL now:
        self.api_url = self.notify_url.format(
            proj=self.project_id,
            hook=self.msghook,
        )
        return

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (self.secure_protocol, self.project_id, self.msghook)

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            'color': 'yes' if self.color else 'no',
            'prefix': 'yes' if self.prefix else 'no',
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return '{schema}://{proj}/{hook}/?{params}'.format(
            schema=self.secure_protocol,
            proj=self.pprint(self.project_id, privacy, safe=''),
            hook=self.pprint(self.msghook, privacy, safe=''),
            params=NotifyNotifico.urlencode(params),
        )

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        wrapper to _send since we can alert more then one channel
        """

        # prepare our headers
        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
        }

        # Prepare our IRC Prefix
        color = ''
        token = ''
        if notify_type == NotifyType.INFO:
            color = NotificoColor.Teal
            token = 'i'

        elif notify_type == NotifyType.SUCCESS:
            color = NotificoColor.LightGreen
            token = '✔'

        elif notify_type == NotifyType.WARNING:
            color = NotificoColor.Orange
            token = '!'

        elif notify_type == NotifyType.FAILURE:
            color = NotificoColor.Red
            token = '✗'

        if self.color:
            # Colors were specified, make sure we capture and correctly
            # allow them to exist inline in the message
            # \g<1> is less ambiguous than \1
            body = re.sub(r'\\x03(\d{0,2})', r'\\x03\g<1>', body)

        else:
            # no colors specified, make sure we strip out any colors found
            # to make the string read-able
            body = re.sub(r'\\x03(\d{1,2}(,[0-9]{1,2})?)?', r'', body)

        # Prepare our payload
        payload = {
            'payload': body if not self.prefix
            else '{}[{}]{} {}{}{}: {}{}'.format(
                # Token [?] at the head
                color if self.color else '',
                token,
                NotificoColor.Reset if self.color else '',
                # App ID
                NotificoFormat.Bold if self.color else '',
                self.app_id,
                NotificoFormat.Reset if self.color else '',
                # Message Body
                body,
                # Reset
                NotificoFormat.Reset if self.color else '',
            ),
        }

        self.logger.debug('Notifico GET URL: %s (cert_verify=%r)' % (
            self.api_url, self.verify_certificate))
        self.logger.debug('Notifico Payload: %s' % str(payload))

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.get(
                self.api_url,
                params=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = \
                    NotifyNotifico.http_response_code_lookup(r.status_code)

                self.logger.warning(
                    'Failed to send Notifico notification: '
                    '{}{}error={}.'.format(
                        status_str,
                        ', ' if status_str else '',
                        r.status_code))

                self.logger.debug('Response Details:\r\n{}'.format(r.content))

                # Return; we're done
                return False

            else:
                self.logger.info('Sent Notifico notification.')

        except requests.RequestException as e:
            self.logger.warning(
                'A Connection error occurred sending Notifico '
                'notification.')
            self.logger.debug('Socket Exception: %s' % str(e))

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The first token is stored in the hostname
        results['project_id'] = NotifyNotifico.unquote(results['host'])

        # Get Message Hook
        try:
            results['msghook'] = NotifyNotifico.split_path(
                results['fullpath'])[0]

        except IndexError:
            results['msghook'] = None

        # Include Color
        results['color'] = \
            parse_bool(results['qsd'].get('color', True))

        # Include Prefix
        results['prefix'] = \
            parse_bool(results['qsd'].get('prefix', True))

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://n.tkte.ch/h/PROJ_ID/MESSAGE_HOOK/
        """

        result = re.match(
            r'^https?://n\.tkte\.ch/h/'
            r'(?P<proj>[0-9]+)/'
            r'(?P<hook>[A-Z0-9]+)/?'
            r'(?P<params>\?.+)?$', url, re.I)

        if result:
            return NotifyNotifico.parse_url(
                '{schema}://{proj}/{hook}/{params}'.format(
                    schema=NotifyNotifico.secure_protocol,
                    proj=result.group('proj'),
                    hook=result.group('hook'),
                    params='' if not result.group('params')
                    else result.group('params')))

        return None



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from email import charset

from .base import NotifyEmail
from .common import (
    AppriseEmailException, EmailMessage, SecureMailMode, SECURE_MODES,
    WebBaseLogin)
from .templates import EMAIL_TEMPLATES

# Globally Default encoding mode set to Quoted Printable.
charset.add_charset('utf-8', charset.QP, charset.QP, 'utf-8')

__all__ = [
    # Reference
    'NotifyEmail',

    # Pretty Good Privacy
    'ApprisePGPController', 'ApprisePGPException',

    # Other
    'AppriseEmailException', 'EmailMessage', 'SecureMailMode', 'SECURE_MODES',
    'WebBaseLogin',

    # Additional entries that may be useful to some developers
    'EMAIL_TEMPLATES', 'PGP_SUPPORT',
]



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import requests
from json import loads

from .base import NotifyBase
from .. import exception
from ..common import NotifyType
from ..utils.parse import parse_list, validate_regex
from ..locale import gettext_lazy as _


class PushSaferSound:
    """
    Defines all of the supported PushSafe sounds
    """
    # Silent
    SILENT = 0
    # Ahem (IM)
    AHEM = 1
    # Applause (Mail)
    APPLAUSE = 2
    # Arrow (Reminder)
    ARROW = 3
    # Baby (SMS)
    BABY = 4
    # Bell (Alarm)
    BELL = 5
    # Bicycle (Alarm2)
    BICYCLE = 6
    # Boing (Alarm3)
    BOING = 7
    # Buzzer (Alarm4)
    BUZZER = 8
    # Camera (Alarm5)
    CAMERA = 9
    # Car Horn (Alarm6)
    CAR_HORN = 10
    # Cash Register (Alarm7)
    CASH_REGISTER = 11
    # Chime (Alarm8)
    CHIME = 12
    # Creaky Door (Alarm9)
    CREAKY_DOOR = 13
    # Cuckoo Clock (Alarm10)
    CUCKOO_CLOCK = 14
    # Disconnect (Call)
    DISCONNECT = 15
    # Dog (Call2)
    DOG = 16
    # Doorbell (Call3)
    DOORBELL = 17
    # Fanfare (Call4)
    FANFARE = 18
    # Gun Shot (Call5)
    GUN_SHOT = 19
    # Honk (Call6)
    HONK = 20
    # Jaw Harp (Call7)
    JAW_HARP = 21
    # Morse (Call8)
    MORSE = 22
    # Electricity (Call9)
    ELECTRICITY = 23
    # Radio Tuner (Call10)
    RADIO_TURNER = 24
    # Sirens
    SIRENS = 25
    # Military Trumpets
    MILITARY_TRUMPETS = 26
    # Ufo
    UFO = 27
    # Whah Whah Whah
    LONG_WHAH = 28
    # Man Saying Goodbye
    GOODBYE = 29
    # Man Saying Hello
    HELLO = 30
    # Man Saying No
    NO = 31
    # Man Saying Ok
    OKAY = 32
    # Man Saying Ooohhhweee
    OOOHHHWEEE = 33
    # Man Saying Warning
    WARNING = 34
    # Man Saying Welcome
    WELCOME = 35
    # Man Saying Yeah
    YEAH = 36
    # Man Saying Yes
    YES = 37
    # Beep short
    BEEP1 = 38
    # Weeeee short
    WEEE = 39
    # Cut in and out short
    CUTINOUT = 40
    # Finger flicking glas short
    FLICK_GLASS = 41
    # Wa Wa Waaaa short
    SHORT_WHAH = 42
    # Laser short
    LASER = 43
    # Wind Chime short
    WIND_CHIME = 44
    # Echo short
    ECHO = 45
    # Zipper short
    ZIPPER = 46
    # HiHat short
    HIHAT = 47
    # Beep 2 short
    BEEP2 = 48
    # Beep 3 short
    BEEP3 = 49
    # Beep 4 short
    BEEP4 = 50
    # The Alarm is armed
    ALARM_ARMED = 51
    # The Alarm is disarmed
    ALARM_DISARMED = 52
    # The Backup is ready
    BACKUP_READY = 53
    # The Door is closed
    DOOR_CLOSED = 54
    # The Door is opend
    DOOR_OPENED = 55
    # The Window is closed
    WINDOW_CLOSED = 56
    # The Window is open
    WINDOW_OPEN = 57
    # The Light is off
    LIGHT_ON = 58
    # The Light is on
    LIGHT_OFF = 59
    # The Doorbell rings
    DOORBELL_RANG = 60


PUSHSAFER_SOUND_MAP = {
    # Device Default,
    'silent': PushSaferSound.SILENT,
    'ahem': PushSaferSound.AHEM,
    'applause': PushSaferSound.APPLAUSE,
    'arrow': PushSaferSound.ARROW,
    'baby': PushSaferSound.BABY,
    'bell': PushSaferSound.BELL,
    'bicycle': PushSaferSound.BICYCLE,
    'bike': PushSaferSound.BICYCLE,
    'boing': PushSaferSound.BOING,
    'buzzer': PushSaferSound.BUZZER,
    'camera': PushSaferSound.CAMERA,
    'carhorn': PushSaferSound.CAR_HORN,
    'horn': PushSaferSound.CAR_HORN,
    'cashregister': PushSaferSound.CASH_REGISTER,
    'chime': PushSaferSound.CHIME,
    'creakydoor': PushSaferSound.CREAKY_DOOR,
    'cuckooclock': PushSaferSound.CUCKOO_CLOCK,
    'cuckoo': PushSaferSound.CUCKOO_CLOCK,
    'disconnect': PushSaferSound.DISCONNECT,
    'dog': PushSaferSound.DOG,
    'doorbell': PushSaferSound.DOORBELL,
    'fanfare': PushSaferSound.FANFARE,
    'gunshot': PushSaferSound.GUN_SHOT,
    'honk': PushSaferSound.HONK,
    'jawharp': PushSaferSound.JAW_HARP,
    'morse': PushSaferSound.MORSE,
    'electric': PushSaferSound.ELECTRICITY,
    'radiotuner': PushSaferSound.RADIO_TURNER,
    'sirens': PushSaferSound.SIRENS,
    'militarytrumpets': PushSaferSound.MILITARY_TRUMPETS,
    'military': PushSaferSound.MILITARY_TRUMPETS,
    'trumpets': PushSaferSound.MILITARY_TRUMPETS,
    'ufo': PushSaferSound.UFO,
    'whahwhah': PushSaferSound.LONG_WHAH,
    'whah': PushSaferSound.SHORT_WHAH,
    'goodye': PushSaferSound.GOODBYE,
    'hello': PushSaferSound.HELLO,
    'no': PushSaferSound.NO,
    'okay': PushSaferSound.OKAY,
    'ok': PushSaferSound.OKAY,
    'ooohhhweee': PushSaferSound.OOOHHHWEEE,
    'warn': PushSaferSound.WARNING,
    'warning': PushSaferSound.WARNING,
    'welcome': PushSaferSound.WELCOME,
    'yeah': PushSaferSound.YEAH,
    'yes': PushSaferSound.YES,
    'beep': PushSaferSound.BEEP1,
    'beep1': PushSaferSound.BEEP1,
    'weee': PushSaferSound.WEEE,
    'wee': PushSaferSound.WEEE,
    'cutinout': PushSaferSound.CUTINOUT,
    'flickglass': PushSaferSound.FLICK_GLASS,
    'laser': PushSaferSound.LASER,
    'windchime': PushSaferSound.WIND_CHIME,
    'echo': PushSaferSound.ECHO,
    'zipper': PushSaferSound.ZIPPER,
    'hihat': PushSaferSound.HIHAT,
    'beep2': PushSaferSound.BEEP2,
    'beep3': PushSaferSound.BEEP3,
    'beep4': PushSaferSound.BEEP4,
    'alarmarmed': PushSaferSound.ALARM_ARMED,
    'armed': PushSaferSound.ALARM_ARMED,
    'alarmdisarmed': PushSaferSound.ALARM_DISARMED,
    'disarmed': PushSaferSound.ALARM_DISARMED,
    'backupready': PushSaferSound.BACKUP_READY,
    'dooropen': PushSaferSound.DOOR_OPENED,
    'dopen': PushSaferSound.DOOR_OPENED,
    'doorclosed': PushSaferSound.DOOR_CLOSED,
    'dclosed': PushSaferSound.DOOR_CLOSED,
    'windowopen': PushSaferSound.WINDOW_OPEN,
    'wopen': PushSaferSound.WINDOW_OPEN,
    'windowclosed': PushSaferSound.WINDOW_CLOSED,
    'wclosed': PushSaferSound.WINDOW_CLOSED,
    'lighton': PushSaferSound.LIGHT_ON,
    'lon': PushSaferSound.LIGHT_ON,
    'lightoff': PushSaferSound.LIGHT_OFF,
    'loff': PushSaferSound.LIGHT_OFF,
    'doorbellrang': PushSaferSound.DOORBELL_RANG,
}


# Priorities
class PushSaferPriority:
    LOW = -2
    MODERATE = -1
    NORMAL = 0
    HIGH = 1
    EMERGENCY = 2


PUSHSAFER_PRIORITIES = (
    PushSaferPriority.LOW,
    PushSaferPriority.MODERATE,
    PushSaferPriority.NORMAL,
    PushSaferPriority.HIGH,
    PushSaferPriority.EMERGENCY,
)

PUSHSAFER_PRIORITY_MAP = {
    # short for 'low'
    'low': PushSaferPriority.LOW,
    # short for 'medium'
    'medium': PushSaferPriority.MODERATE,
    # short for 'normal'
    'normal': PushSaferPriority.NORMAL,
    # short for 'high'
    'high': PushSaferPriority.HIGH,
    # short for 'emergency'
    'emergency': PushSaferPriority.EMERGENCY,
}

# Identify the priority ou want to designate as the fall back
DEFAULT_PRIORITY = "normal"


# Vibrations
class PushSaferVibration:
    """
    Defines the acceptable vibration settings for notification
    """
    # x1
    LOW = 1
    # x2
    NORMAL = 2
    # x3
    HIGH = 3


# Identify all of the vibrations in one place
PUSHSAFER_VIBRATIONS = (
    PushSaferVibration.LOW,
    PushSaferVibration.NORMAL,
    PushSaferVibration.HIGH,
)

# At this time, the following pictures can be attached to each notification
# at one time. When more are supported, just add their argument below
PICTURE_PARAMETER = (
    'p',
    'p2',
    'p3',
)


# Flag used as a placeholder to sending to all devices
PUSHSAFER_SEND_TO_ALL = 'a'


class NotifyPushSafer(NotifyBase):
    """
    A wrapper for PushSafer Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'Pushsafer'

    # The services URL
    service_url = 'https://www.pushsafer.com/'

    # The default insecure protocol
    protocol = 'psafer'

    # The default secure protocol
    secure_protocol = 'psafers'

    # Support attachments
    attachment_support = True

    # Number of requests to a allow per second
    request_rate_per_sec = 1.2

    # The icon ID of 25 looks like a megaphone
    default_pushsafer_icon = 25

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_pushsafer'

    # Defines the hostname to post content to; since this service supports
    # both insecure and secure methods, we set the {schema} just before we
    # post the message upstream.
    notify_url = '{schema}://www.pushsafer.com/api'

    # Define object templates
    templates = (
        '{schema}://{privatekey}',
        '{schema}://{privatekey}/{targets}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'privatekey': {
            'name': _('Private Key'),
            'type': 'string',
            'private': True,
            'required': True,
        },
        'target_device': {
            'name': _('Target Device'),
            'type': 'string',
            'map_to': 'targets',
        },
        'target_email': {
            'name': _('Target Email'),
            'type': 'string',
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'priority': {
            'name': _('Priority'),
            'type': 'choice:int',
            'values': PUSHSAFER_PRIORITIES,
        },
        'sound': {
            'name': _('Sound'),
            'type': 'choice:string',
            'values': PUSHSAFER_SOUND_MAP,
        },
        'vibration': {
            'name': _('Vibration'),
            'type': 'choice:int',
            'values': PUSHSAFER_VIBRATIONS,
        },
        'to': {
            'alias_of': 'targets',
        },
    })

    def __init__(self, privatekey, targets=None, priority=None, sound=None,
                 vibration=None, **kwargs):
        """
        Initialize PushSafer Object
        """
        super().__init__(**kwargs)

        #
        # Priority
        #
        try:
            # Acquire our priority if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.priority = int(priority)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.priority = None

        except ValueError:
            # Input is a string; attempt to get the lookup from our
            # priority mapping
            priority = priority.lower().strip()

            # This little bit of black magic allows us to match against
            # low, lo, l (for low);
            # normal, norma, norm, nor, no, n (for normal)
            # ... etc
            match = next((key for key in PUSHSAFER_PRIORITY_MAP.keys()
                         if key.startswith(priority)), None) \
                if priority else None

            # Now test to see if we got a match
            if not match:
                msg = 'An invalid PushSafer priority ' \
                      '({}) was specified.'.format(priority)
                self.logger.warning(msg)
                raise TypeError(msg)

            # store our successfully looked up priority
            self.priority = PUSHSAFER_PRIORITY_MAP[match]

        if self.priority is not None and \
                self.priority not in PUSHSAFER_PRIORITY_MAP.values():
            msg = 'An invalid PushSafer priority ' \
                  '({}) was specified.'.format(priority)
            self.logger.warning(msg)
            raise TypeError(msg)

        #
        # Sound
        #
        try:
            # Acquire our sound if we can:
            #  - We accept both the integer form as well as a string
            #    representation
            self.sound = int(sound)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.sound = None

        except ValueError:
            # Input is a string; attempt to get the lookup from our
            # sound mapping
            sound = sound.lower().strip()

            # This little bit of black magic allows us to match against
            # against multiple versions of the same string
            # ... etc
            match = next((key for key in PUSHSAFER_SOUND_MAP.keys()
                         if key.startswith(sound)), None) \
                if sound else None

            # Now test to see if we got a match
            if not match:
                msg = 'An invalid PushSafer sound ' \
                      '({}) was specified.'.format(sound)
                self.logger.warning(msg)
                raise TypeError(msg)

            # store our successfully looked up sound
            self.sound = PUSHSAFER_SOUND_MAP[match]

        if self.sound is not None and \
                self.sound not in PUSHSAFER_SOUND_MAP.values():
            msg = 'An invalid PushSafer sound ' \
                  '({}) was specified.'.format(sound)
            self.logger.warning(msg)
            raise TypeError(msg)

        #
        # Vibration
        #
        try:
            # Use defined integer as is if defined, no further error checking
            # is performed
            self.vibration = int(vibration)

        except TypeError:
            # NoneType means use Default; this is an okay exception
            self.vibration = None

        except ValueError:
            msg = 'An invalid PushSafer vibration ' \
                  '({}) was specified.'.format(vibration)
            self.logger.warning(msg)
            raise TypeError(msg)

        if self.vibration and self.vibration not in PUSHSAFER_VIBRATIONS:
            msg = 'An invalid PushSafer vibration ' \
                  '({}) was specified.'.format(vibration)
            self.logger.warning(msg)
            raise TypeError(msg)

        #
        # Private Key (associated with project)
        #
        self.privatekey = validate_regex(privatekey)
        if not self.privatekey:
            msg = 'An invalid PushSafer Private Key ' \
                  '({}) was specified.'.format(privatekey)
            self.logger.warning(msg)
            raise TypeError(msg)

        self.targets = parse_list(targets)
        if len(self.targets) == 0:
            self.targets = (PUSHSAFER_SEND_TO_ALL, )

        return

    def send(self, body, title='', notify_type=NotifyType.INFO, attach=None,
             **kwargs):
        """
        Perform PushSafer Notification
        """

        # error tracking (used for function return)
        has_error = False

        # Initialize our list of attachments
        attachments = []

        if attach and self.attachment_support:
            # We need to upload our payload first so that we can source it
            # in remaining messages
            for no, attachment in enumerate(attach, start=1):
                # prepare payload
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        'Could not access PushSafer attachment {}.'.format(
                            attachment.url(privacy=True)))
                    return False

                if not attachment.mimetype.startswith('image/'):
                    # Attachment not supported; continue peacefully
                    self.logger.debug(
                        'Ignoring unsupported PushSafer attachment {}.'.format(
                            attachment.url(privacy=True)))
                    continue

                self.logger.debug(
                    'Posting PushSafer attachment {}'.format(
                        attachment.url(privacy=True)))

                try:
                    # Output must be in a DataURL format (that's what
                    # PushSafer calls it):
                    attachments.append((
                        attachment.name
                        if attachment.name else f'file{no:03}.dat',
                        'data:{};base64,{}'.format(
                            attachment.mimetype,
                            attachment.base64(),
                        )
                    ))

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        'Could not access PushSafer attachment {}.'.format(
                            attachment.url(privacy=True)))
                    return False

                self.logger.debug(
                    'Appending PushSafer attachment {}'.format(
                        attachment.url(privacy=True)))

        # Create a copy of the targets list
        targets = list(self.targets)
        while len(targets):
            recipient = targets.pop(0)

            # prepare payload
            payload = {
                't': title,
                'm': body,
                # Our default icon to use
                'i': self.default_pushsafer_icon,
                # Notification Color
                'c': self.color(notify_type),
                # Target Recipient
                'd': recipient,
            }

            if self.sound is not None:
                # Only apply sound setting if it was specified
                payload['s'] = str(self.sound)

            if self.vibration is not None:
                # Only apply vibration setting
                payload['v'] = str(self.vibration)

            if not attachments:
                okay, response = self._send(payload)
                if not okay:
                    has_error = True
                    continue

                self.logger.info(
                    'Sent PushSafer notification to "%s".' % (recipient))

            else:
                # Create a copy of our payload object
                _payload = payload.copy()

                for idx in range(
                        0, len(attachments), len(PICTURE_PARAMETER)):
                    # Send our attachments to our same user (already prepared
                    # as our payload object)
                    for c, attachment in enumerate(
                            attachments[idx:idx + len(PICTURE_PARAMETER)]):

                        # Get our attachment information
                        filename, dataurl = attachment
                        _payload.update({PICTURE_PARAMETER[c]: dataurl})

                        self.logger.debug(
                            'Added attachment (%s) to "%s".' % (
                                filename, recipient))

                    okay, response = self._send(_payload)
                    if not okay:
                        has_error = True
                        continue

                    self.logger.info(
                        'Sent PushSafer attachment (%s) to "%s".' % (
                            filename, recipient))

                    # More then the maximum messages shouldn't cause all of
                    # the text to loop on future iterations
                    _payload = payload.copy()
                    _payload['t'] = ''
                    _payload['m'] = '...'

        return not has_error

    def _send(self, payload, **kwargs):
        """
        Wrapper to the requests (post) object
        """

        headers = {
            'User-Agent': self.app_id,
        }

        # Prepare the notification URL to post to
        notify_url = self.notify_url.format(
            schema='https' if self.secure else 'http'
        )

        # Store the payload key
        payload['k'] = self.privatekey

        self.logger.debug('PushSafer POST URL: %s (cert_verify=%r)' % (
            notify_url, self.verify_certificate,
        ))
        self.logger.debug('PushSafer Payload: %s' % str(payload))

        # Always call throttle before any remote server i/o is made
        self.throttle()

        # Default response type
        response = None

        # Initialize our Pushsafer expected responses
        _code = None
        _str = 'Unknown'

        try:
            # Open our attachment path if required:
            r = requests.post(
                notify_url,
                data=payload,
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                response = loads(r.content)
                _code = response.get('status')
                _str = response.get('success', _str) \
                    if _code == 1 else response.get('error', _str)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None

                # Fall back to the existing unparsed value
                response = r.content

            if r.status_code not in (
                    requests.codes.ok, requests.codes.no_content):
                # We had a problem
                status_str = \
                    NotifyPushSafer.http_response_code_lookup(
                        r.status_code)

                self.logger.warning(
                    'Failed to deliver payload to PushSafer:'
                    '{}{}error={}.'.format(
                        status_str,
                        ', ' if status_str else '',
                        r.status_code))

                self.logger.debug(
                    'Response Details:\r\n{}'.format(r.content))

                return False, response

            elif _code != 1:
                # It's a bit backwards, but:
                #    1 is returned if we succeed
                #    0 is returned if we fail
                self.logger.warning(
                    'Failed to deliver payload to PushSafer;'
                    ' error={}.'.format(_str))

                self.logger.debug(
                    'Response Details:\r\n{}'.format(r.content))

                return False, response

            # otherwise we were successful
            return True, response

        except requests.RequestException as e:
            self.logger.warning(
                'A Connection error occurred communicating with PushSafer.')
            self.logger.debug('Socket Exception: %s' % str(e))

            return False, response

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.privatekey,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.priority is not None:
            # Store our priority; but only if it was specified
            params['priority'] = \
                next((key for key, value in PUSHSAFER_PRIORITY_MAP.items()
                      if value == self.priority),
                     DEFAULT_PRIORITY)  # pragma: no cover

        if self.sound is not None:
            # Store our sound; but only if it was specified
            params['sound'] = \
                next((key for key, value in PUSHSAFER_SOUND_MAP.items()
                      if value == self.sound), '')  # pragma: no cover

        if self.vibration is not None:
            # Store our vibration; but only if it was specified
            params['vibration'] = str(self.vibration)

        targets = '/'.join([NotifyPushSafer.quote(x) for x in self.targets])
        if targets == PUSHSAFER_SEND_TO_ALL:
            # keyword is reserved for internal usage only; it's safe to remove
            # it from the recipients list
            targets = ''

        return '{schema}://{privatekey}/{targets}?{params}'.format(
            schema=self.secure_protocol if self.secure else self.protocol,
            privatekey=self.pprint(self.privatekey, privacy, safe=''),
            targets=targets,
            params=NotifyPushSafer.urlencode(params))

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Fetch our targets
        results['targets'] = \
            NotifyPushSafer.split_path(results['fullpath'])

        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'] += \
                NotifyPushSafer.parse_list(results['qsd']['to'])

        # Setup the token; we store it in Private Key for global
        # plugin consistency with naming conventions
        results['privatekey'] = NotifyPushSafer.unquote(results['host'])

        if 'priority' in results['qsd'] and len(results['qsd']['priority']):
            results['priority'] = \
                NotifyPushSafer.unquote(results['qsd']['priority'])

        if 'sound' in results['qsd'] and len(results['qsd']['sound']):
            results['sound'] = \
                NotifyPushSafer.unquote(results['qsd']['sound'])

        if 'vibration' in results['qsd'] and len(results['qsd']['vibration']):
            results['vibration'] = \
                NotifyPushSafer.unquote(results['qsd']['vibration'])

        return results



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://messagebird.com if you don't already have one
#
# Get your (apikey) and api example from the dashboard here:
#   - https://dashboard.messagebird.com/en/user/index
#

import requests

from .base import NotifyBase
from ..common import NotifyType
from ..utils.parse import is_phone_no, parse_phone_no, validate_regex
from ..locale import gettext_lazy as _


class NotifyMessageBird(NotifyBase):
    """
    A wrapper for MessageBird Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'MessageBird'

    # The services URL
    service_url = 'https://messagebird.com'

    # The default protocol
    secure_protocol = 'msgbird'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_messagebird'

    # MessageBird uses the http protocol with JSON requests
    notify_url = 'https://rest.messagebird.com/messages'

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        '{schema}://{apikey}/{source}',
        '{schema}://{apikey}/{source}/{targets}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'apikey': {
            'name': _('API Key'),
            'type': 'string',
            'required': True,
            'private': True,
            'regex': (r'^[a-z0-9]{25}$', 'i'),
        },
        'source': {
            'name': _('Source Phone No'),
            'type': 'string',
            'prefix': '+',
            'required': True,
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
        },
        'target_phone': {
            'name': _('Target Phone No'),
            'type': 'string',
            'prefix': '+',
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
        }
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'to': {
            'alias_of': 'targets',
        },
        'from': {
            'alias_of': 'source',
        },
    })

    def __init__(self, apikey, source, targets=None, **kwargs):
        """
        Initialize MessageBird Object
        """
        super().__init__(**kwargs)

        # API Key (associated with project)
        self.apikey = validate_regex(
            apikey, *self.template_tokens['apikey']['regex'])
        if not self.apikey:
            msg = 'An invalid MessageBird API Key ' \
                  '({}) was specified.'.format(apikey)
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_phone_no(source)
        if not result:
            msg = 'The MessageBird source specified ({}) is invalid.'\
                .format(source)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our source
        self.source = result['full']

        # Parse our targets
        self.targets = list()

        targets = parse_phone_no(targets)
        if not targets:
            # No sources specified, use our own phone no
            self.targets.append(self.source)
            return

        # otherwise, store all of our target numbers
        for target in targets:
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    'Dropped invalid phone # '
                    '({}) specified.'.format(target),
                )
                continue

            # store valid phone number
            self.targets.append(result['full'])

        return

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform MessageBird Notification
        """

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning('There were no MessageBird targets to notify.')
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'AccessKey {}'.format(self.apikey),
        }

        # Prepare our payload
        payload = {
            'originator': '+{}'.format(self.source),
            'recipients': None,
            'body': body,

        }

        # Create a copy of the targets list
        targets = list(self.targets)

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload['recipients'] = '+{}'.format(target)

            # Some Debug Logging
            self.logger.debug(
                'MessageBird POST URL: {} (cert_verify={})'.format(
                    self.notify_url, self.verify_certificate))
            self.logger.debug('MessageBird Payload: {}' .format(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # Sample output of a successful transmission
                # {
                #   "originator": "+15553338888",
                #   "body": "test",
                #   "direction": "mt",
                #   "mclass": 1,
                #   "reference": null,
                #   "createdDatetime": "2019-08-22T01:32:18+00:00",
                #   "recipients": {
                #     "totalCount": 1,
                #     "totalSentCount": 1,
                #     "totalDeliveredCount": 0,
                #     "totalDeliveryFailedCount": 0,
                #     "items": [
                #       {
                #         "status": "sent",
                #         "statusDatetime": "2019-08-22T01:32:18+00:00",
                #         "recipient": 15553338888,
                #         "messagePartCount": 1
                #       }
                #     ]
                #   },
                #   "validity": null,
                #   "gateway": 10,
                #   "typeDetails": {},
                #   "href": "https://rest.messagebird.com/messages/\
                #       b5d424244a5b4fd0b5b5728bccaafc23",
                #   "datacoding": "plain",
                #   "scheduledDatetime": null,
                #   "type": "sms",
                #   "id": "b5d424244a5b4fd0b5b5728bccaafc23"
                # }

                if r.status_code not in (
                        requests.codes.ok, requests.codes.created):
                    # We had a problem
                    status_str = \
                        NotifyMessageBird.http_response_code_lookup(
                            r.status_code)

                    self.logger.warning(
                        'Failed to send MessageBird notification to {}: '
                        '{}{}error={}.'.format(
                            ','.join(target),
                            status_str,
                            ', ' if status_str else '',
                            r.status_code))

                    self.logger.debug(
                        'Response Details:\r\n{}'.format(r.content))

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        'Sent MessageBird notification to {}.'.format(target))

            except requests.RequestException as e:
                self.logger.warning(
                    'A Connection error occurred sending MessageBird:%s ' % (
                        target) + 'notification.'
                )
                self.logger.debug('Socket Exception: %s' % str(e))

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (self.secure_protocol, self.apikey, self.source)

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return '{schema}://{apikey}/{source}/{targets}/?{params}'.format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=''),
            source=self.source,
            targets='/'.join(
                [NotifyMessageBird.quote(x, safe='') for x in self.targets]),
            params=NotifyMessageBird.urlencode(params))

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results['targets'] = NotifyMessageBird.split_path(results['fullpath'])

        try:
            # The first path entry is the source/originator
            results['source'] = results['targets'].pop(0)

        except IndexError:
            # No path specified... this URL is potentially un-parseable; we can
            # hope for a from= entry
            results['source'] = None

        # The hostname is our authentication key
        results['apikey'] = NotifyMessageBird.unquote(results['host'])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'] += \
                NotifyMessageBird.parse_phone_no(results['qsd']['to'])

        if 'from' in results['qsd'] and len(results['qsd']['from']):
            results['source'] = \
                NotifyMessageBird.unquote(results['qsd']['from'])

        return results



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2023, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://messagebird.com if you don't already have one
#
# Get your auth_id and auth token from the dashboard here:
#   - https://console.plivo.com/dashboard/
#

import requests

from json import dumps
from .base import NotifyBase
from ..common import NotifyType
from ..utils.parse import (
    parse_bool, is_phone_no, parse_phone_no, validate_regex)
from ..locale import gettext_lazy as _


class NotifyPlivo(NotifyBase):
    """
    A wrapper for Plivo Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'Plivo'

    # The services URL
    service_url = 'https://plivo.com'

    # The default protocol
    secure_protocol = 'plivo'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_plivo'

    # Plivo uses the http protocol with JSON requests
    notify_url = 'https://api.plivo.com/v1/Account/{auth_id}/Message/'

    # The maximum number of messages that can be sent in a single batch
    default_batch_size = 20

    # The maximum length of the body
    body_maxlen = 140

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        '{schema}://{auth_id}@{token}/{source}',
        '{schema}://{auth_id}@{token}/{source}/{targets}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'auth_id': {
            'name': _('Auth ID'),
            'type': 'string',
            'required': True,
            'regex': (r'^[a-z0-9]{20,30}$', 'i'),
        },
        'token': {
            'name': _('Auth Token'),
            'type': 'string',
            'required': True,
            'regex': (r'^[a-z0-9]{30,50}$', 'i'),
        },
        'source': {
            'name': _('Source Phone No'),
            'type': 'string',
            'prefix': '+',
            'required': True,
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
        },
        'target_phone': {
            'name': _('Target Phone No'),
            'type': 'string',
            'prefix': '+',
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
        }
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'to': {
            'alias_of': 'targets',
        },
        'from': {
            'alias_of': 'source',
        },
        'token': {
            'alias_of': 'token',
        },
        'id': {
            'alias_of': 'auth_id',
        },
        'batch': {
            'name': _('Batch Mode'),
            'type': 'bool',
            'default': False,
        },
    })

    def __init__(self, auth_id, token, source, targets=None, batch=None,
                 **kwargs):
        """
        Initialize Plivo Object
        """
        super(NotifyPlivo, self).__init__(**kwargs)

        self.auth_id = validate_regex(
            auth_id, *self.template_tokens['auth_id']['regex'])
        if not self.auth_id:
            msg = 'The Plivo authentication ID specified ({}) is ' \
                'invalid.'.format(auth_id)
            self.logger.warning(msg)
            raise TypeError(msg)

        self.token = validate_regex(
            token, *self.template_tokens['token']['regex'])
        if not self.token:
            msg = 'The Plivo authentication token specified ({}) is ' \
                'invalid.'.format(token)
            self.logger.warning(msg)
            raise TypeError(msg)

        result = is_phone_no(source)
        if not result:
            msg = 'The Plivo source specified ({}) is invalid.'\
                .format(source)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our source; enforce E.164 format
        self.source = f'+{result["full"]}'

        # Parse our targets
        self.targets = list()

        if targets:
            for target in parse_phone_no(targets):
                # Validate targets and drop bad ones:
                result = is_phone_no(target)
                if result:
                    # store valid phone number; enforce E.164 format
                    self.targets.append(f'+{result["full"]}')
                    continue

                self.logger.warning(
                    'Dropped invalid phone # '
                    '({}) specified.'.format(target),
                )
        else:
            # No sources specified, use our own phone no
            self.targets.append(self.source)

        # Set batch
        self.batch = batch if batch is not None \
            else self.template_args['batch']['default']

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform Plivo Notification
        """

        if not self.targets:
            # There were no services to notify
            self.logger.warning(
                'There were no Plivo targets to notify.')
            return False

        # Initialize our has_error flag
        has_error = False

        # Prepare our headers
        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/json',
        }

        # Prepare our authentication
        auth = (self.auth_id, self.token)

        # Prepare our payload
        payload = {
            'src': self.source,
            'dst': None,
            'text': body,
        }

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        for index in range(0, len(self.targets), batch_size):
            # Prepare our phone no (< delimits more then one)
            payload['recipients'] = \
                ','.join(self.targets[index:index + batch_size])

            # Some Debug Logging
            self.logger.debug(
                'Plivo POST URL: {} (cert_verify={})'.format(
                    self.notify_url, self.verify_certificate))
            self.logger.debug('Plivo Payload: {}' .format(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code not in (
                        requests.codes.ok, requests.codes.accepted):
                    # We had a problem
                    status_str = \
                        NotifyPlivo.http_response_code_lookup(
                            r.status_code)

                    self.logger.warning(
                        'Failed to send {} Plivo notification{}: '
                        '{}{}error={}.'.format(
                            len(self.targets[index:index + batch_size]),
                            ' to {}'.format(self.targets[index])
                            if batch_size == 1 else '(s)',
                            status_str,
                            ', ' if status_str else '',
                            r.status_code))

                    self.logger.debug(
                        'Response Details:\r\n{}'.format(r.content))

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        'Send {} Plivo notification{}'.format(
                            len(self.targets[index:index + batch_size]),
                            ' to {}'.format(self.targets[index])
                            if batch_size == 1 else '(s)',
                        ))

            except requests.RequestException as e:
                self.logger.warning(
                    'A Connection error occured sending Plivo:%s ' % (
                        self.targets) + 'notification.'
                )
                self.logger.debug('Socket Exception: %s' % str(e))

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.auth_id, self.token, self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any arguments set
        params = {
            'batch': 'yes' if self.batch else 'no',
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return '{schema}://{auth_id}@{token}/{source}/' \
            '{targets}/?{params}'.format(
                schema=self.secure_protocol,
                auth_id=self.pprint(self.auth_id, privacy, safe=''),
                token=self.pprint(self.token, privacy, safe=''),
                source=self.source,
                targets='/'.join(
                    [NotifyPlivo.quote(x, safe='+') for x in self.targets]),
                params=NotifyPlivo.urlencode(params))

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        #
        # Factor batch into calculation
        #
        return len(self.targets) if self.targets else 1

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to substantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The Auth ID is in the username field
        if 'id' in results['qsd'] and len(results['qsd']['id']):
            results['auth_id'] = NotifyPlivo.unquote(results['qsd']['id'])

        else:
            results['auth_id'] = NotifyPlivo.unquote(results['user'])

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results['targets'] = NotifyPlivo.split_path(results['fullpath'])
        if 'token' in results['qsd'] and len(results['qsd']['token']):
            # Store token
            results['token'] = NotifyPlivo.unquote(results['qsd']['token'])

            # go ahead and put the host entry in the targets list
            if results['host']:
                results['targets'].insert(
                    0, NotifyPlivo.unquote(results['host']))

        else:
            # The hostname is our authentication key
            results['token'] = NotifyPlivo.unquote(results['host'])

        if 'from' in results['qsd'] and len(results['qsd']['from']):
            results['source'] = \
                NotifyPlivo.unquote(results['qsd']['from'])

        else:
            try:
                # The first path entry is the source/originator
                results['source'] = results['targets'].pop(0)

            except IndexError:
                # No source specified...
                results['source'] = None
                pass

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'] += \
                NotifyPlivo.parse_phone_no(results['qsd']['to'])

        # Get Batch Mode Flag
        results['batch'] = \
            parse_bool(results['qsd'].get(
                'batch', NotifyPlivo.template_args['batch']['default']))

        return results



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# PAHO MQTT Documentation:
#  https://www.eclipse.org/paho/index.php?page=clients/python/docs/index.php
#
# Looking at the PAHO MQTT Source can help shed light on what's going on too
# as their inline documentation is pretty good!
#   https://github.com/eclipse/paho.mqtt.python\
#           /blob/master/src/paho/mqtt/client.py
import ssl
import re
from time import sleep
from datetime import datetime
from os.path import isfile
from .base import NotifyBase
from ..url import PrivacyMode
from ..common import NotifyType
from ..utils.parse import parse_list, parse_bool
from ..locale import gettext_lazy as _

# Default our global support flag
NOTIFY_MQTT_SUPPORT_ENABLED = False

try:
    # 3rd party modules
    import paho.mqtt.client as mqtt

    # We're good to go!
    NOTIFY_MQTT_SUPPORT_ENABLED = True

    MQTT_PROTOCOL_MAP = {
        # v3.1.1
        "311": mqtt.MQTTv311,
        # v3.1
        "31": mqtt.MQTTv31,
        # v5.0
        "5": mqtt.MQTTv5,
        # v5.0 (alias)
        "50": mqtt.MQTTv5,
    }

except ImportError:
    # No problem; we just simply can't support this plugin because we're
    # either using Linux, or simply do not have pywin32 installed.
    MQTT_PROTOCOL_MAP = {}

# A lookup map for relaying version to user
HUMAN_MQTT_PROTOCOL_MAP = {
    "v3.1.1": "311",
    "v3.1": "31",
    "v5.0": "5",
}


class NotifyMQTT(NotifyBase):
    """
    A wrapper for MQTT Notifications
    """

    # Set our global enabled flag
    enabled = NOTIFY_MQTT_SUPPORT_ENABLED

    requirements = {
        # Define our required packaging in order to work
        'packages_required': 'paho-mqtt != 2.0.*'
    }

    # The default descriptive name associated with the Notification
    service_name = 'MQTT Notification'

    # The default protocol
    protocol = 'mqtt'

    # Secure protocol
    secure_protocol = 'mqtts'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_mqtt'

    # MQTT does not have a title
    title_maxlen = 0

    # The maximum length a body can be set to
    body_maxlen = 268435455

    # Use a throttle; but it doesn't need to be so strict since most
    # MQTT server hostings can handle the small bursts of packets and are
    # locally hosted anyway
    request_rate_per_sec = 0.5

    # Port Defaults (unless otherwise specified)
    mqtt_insecure_port = 1883

    # The default secure port to use (if mqtts://)
    mqtt_secure_port = 8883

    # The default mqtt keepalive value
    mqtt_keepalive = 30

    # The default mqtt transport
    mqtt_transport = "tcp"

    # The number of seconds to wait for a publish to occur at before
    # checking to see if it's been sent yet.
    mqtt_block_time_sec = 0.2

    # Set the maximum number of messages with QoS>0 that can be part way
    # through their network flow at once.
    mqtt_inflight_messages = 200

    # Define object templates
    templates = (
        '{schema}://{user}@{host}/{topic}',
        '{schema}://{user}@{host}:{port}/{topic}',
        '{schema}://{user}:{password}@{host}/{topic}',
        '{schema}://{user}:{password}@{host}:{port}/{topic}',
    )

    template_tokens = dict(NotifyBase.template_tokens, **{
        'host': {
            'name': _('Hostname'),
            'type': 'string',
            'required': True,
        },
        'port': {
            'name': _('Port'),
            'type': 'int',
            'min': 1,
            'max': 65535,
        },
        'user': {
            'name': _('User Name'),
            'type': 'string',
            'required': True,
        },
        'password': {
            'name': _('Password'),
            'type': 'string',
            'private': True,
            'required': True,
        },
        'topic': {
            'name': _('Target Queue'),
            'type': 'string',
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'to': {
            'alias_of': 'targets',
        },
        'qos': {
            'name': _('QOS'),
            'type': 'int',
            'default': 0,
            'min': 0,
            'max': 2,
        },
        'version': {
            'name': _('Version'),
            'type': 'choice:string',
            'values': HUMAN_MQTT_PROTOCOL_MAP,
            'default': "v3.1.1",
        },
        'client_id': {
            'name': _('Client ID'),
            'type': 'string',
        },
        'session': {
            'name': _('Use Session'),
            'type': 'bool',
            'default': False,
        },
        'retain': {
            'name': _('Retain Messages'),
            'type': 'bool',
            'default': False,
        },
    })

    def __init__(self, targets=None, version=None, qos=None,
                 client_id=None, session=None, retain=None, **kwargs):
        """
        Initialize MQTT Object
        """

        super().__init__(**kwargs)

        # Initialize topics
        self.topics = parse_list(targets)

        if version is None:
            self.version = self.template_args['version']['default']
        else:
            self.version = version

        # Save our client id if specified
        self.client_id = client_id

        # Maintain our session (associated with our user id if set)
        self.session = self.template_args['session']['default'] \
            if session is None or not self.client_id \
            else parse_bool(session)

        # Our Retain Message Flag
        self.retain = self.template_args['retain']['default'] \
            if retain is None else parse_bool(retain)

        # Set up our Quality of Service (QoS)
        try:
            self.qos = self.template_args['qos']['default'] \
                if qos is None else int(qos)

            if self.qos < self.template_args['qos']['min'] \
                    or self.qos > self.template_args['qos']['max']:
                # Let error get handle on exceptio higher up
                raise ValueError("")

        except (ValueError, TypeError):
            msg = 'An invalid MQTT QOS ({}) was specified.'.format(qos)
            self.logger.warning(msg)
            raise TypeError(msg)

        if not self.port:
            # Assign port (if not otherwise set)
            self.port = self.mqtt_secure_port \
                if self.secure else self.mqtt_insecure_port

        self.ca_certs = None
        if self.secure:
            # verify SSL key or abort
            # TODO: There is no error reporting or aborting here?
            #       It could be useful to inform the user _where_ Apprise
            #       tried to find the root CA certificates file.
            self.ca_certs = next(
                (cert for cert in self.CA_CERTIFICATE_FILE_LOCATIONS
                 if isfile(cert)), None)

        # Set up our MQTT Publisher
        try:
            # Get our protocol
            self.mqtt_protocol = \
                MQTT_PROTOCOL_MAP[re.sub(r'[^0-9]+', '', self.version)]

        except (KeyError):
            msg = 'An invalid MQTT Protocol version ' \
                '({}) was specified.'.format(version)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Our MQTT Client Object
        self.client = mqtt.Client(
            client_id=self.client_id,
            clean_session=not self.session, userdata=None,
            protocol=self.mqtt_protocol, transport=self.mqtt_transport,
        )

        # Our maximum number of in-flight messages
        self.client.max_inflight_messages_set(self.mqtt_inflight_messages)

        # Toggled to False once our connection has been established at least
        # once
        self.__initial_connect = True

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform MQTT Notification
        """

        if len(self.topics) == 0:
            # There were no services to notify
            self.logger.warning('There were no MQTT topics to notify.')
            return False

        # For logging:
        url = '{host}:{port}'.format(host=self.host, port=self.port)

        try:
            if self.__initial_connect:
                # Our initial connection
                if self.user:
                    self.client.username_pw_set(
                        self.user, password=self.password)

                if self.secure:
                    if self.ca_certs is None:
                        self.logger.error(
                            'MQTT secure communication can not be verified, '
                            'CA certificates file missing')
                        return False

                    self.client.tls_set(
                        ca_certs=self.ca_certs, certfile=None, keyfile=None,
                        cert_reqs=ssl.CERT_REQUIRED,
                        tls_version=ssl.PROTOCOL_TLS,
                        ciphers=None)

                    # Set our TLS Verify Flag
                    self.client.tls_insecure_set(not self.verify_certificate)

                # Establish our connection
                if self.client.connect(
                        self.host, port=self.port,
                        keepalive=self.mqtt_keepalive) \
                        != mqtt.MQTT_ERR_SUCCESS:
                    self.logger.warning(
                        'An MQTT connection could not be established for {}'.
                        format(url))
                    return False

                # Start our client loop
                self.client.loop_start()

                # Throttle our start otherwise the starting handshaking doesnt
                # work. I'm not sure if this is a bug or not, but with qos=0,
                # and without this sleep(), the messages randomly fails to be
                # delivered.
                sleep(0.01)

                # Toggle our flag since we never need to enter this area again
                self.__initial_connect = False

            # Create a copy of the subreddits list
            topics = list(self.topics)

            has_error = False
            while len(topics) > 0 and not has_error:
                # Retrieve our subreddit
                topic = topics.pop()

                # For logging:
                url = '{host}:{port}/{topic}'.format(
                    host=self.host,
                    port=self.port,
                    topic=topic)

                # Always call throttle before any remote server i/o is made
                self.throttle()

                # handle a re-connection
                if not self.client.is_connected() and \
                        self.client.reconnect() != mqtt.MQTT_ERR_SUCCESS:
                    self.logger.warning(
                        'An MQTT connection could not be sustained for {}'.
                        format(url))
                    has_error = True
                    break

                # Some Debug Logging
                self.logger.debug('MQTT POST URL: {} (cert_verify={})'.format(
                    url, self.verify_certificate))
                self.logger.debug('MQTT Payload: %s' % str(body))

                result = self.client.publish(
                    topic, payload=body, qos=self.qos, retain=self.retain)

                if result.rc != mqtt.MQTT_ERR_SUCCESS:
                    # Toggle our status
                    self.logger.warning(
                        'An error (rc={}) occured when sending MQTT to {}'.
                        format(result.rc, url))
                    has_error = True
                    break

                elif not result.is_published():
                    self.logger.debug(
                        'Blocking until MQTT payload is published...')
                    reference = datetime.now()
                    while not has_error and not result.is_published():
                        # Throttle
                        sleep(self.mqtt_block_time_sec)

                        # Our own throttle so we can abort eventually....
                        elapsed = (datetime.now() - reference).total_seconds()
                        if elapsed >= self.socket_read_timeout:
                            self.logger.warning(
                                'The MQTT message could not be delivered')
                            has_error = True

                # if we reach here; we're at the bottom of our loop
                # we loop around and do the next topic now

        except ConnectionError as e:
            self.logger.warning(
                'MQTT Connection Error received from {}'.format(url))
            self.logger.debug('Socket Exception: %s' % str(e))
            return False

        except ssl.CertificateError as e:
            self.logger.warning(
                'MQTT SSL Certificate Error received from {}'.format(url))
            self.logger.debug('Socket Exception: %s' % str(e))
            return False

        except ValueError as e:
            # ValueError's are thrown from publish() call if there is a problem
            self.logger.warning(
                'MQTT Publishing error received: from {}'.format(url))
            self.logger.debug('Socket Exception: %s' % str(e))
            return False

        if not has_error:
            # Verbal notice
            self.logger.info('Sent MQTT notification')

        return not has_error

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user, self.password, self.host,
            self.port if self.port else (
                self.mqtt_secure_port if self.secure
                else self.mqtt_insecure_port),
            self.fullpath.rstrip('/'),
            self.client_id,
        )

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            'version': self.version,
            'qos': str(self.qos),
            'session': 'yes' if self.session else 'no',
            'retain': 'yes' if self.retain else 'no',
        }

        if self.client_id:
            # Our client id is set if specified
            params['client_id'] = self.client_id

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ''
        if self.user and self.password:
            auth = '{user}:{password}@'.format(
                user=NotifyMQTT.quote(self.user, safe=''),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=''),
            )
        elif self.user:
            auth = '{user}@'.format(
                user=NotifyMQTT.quote(self.user, safe=''),
            )

        default_port = self.mqtt_secure_port \
            if self.secure else self.mqtt_insecure_port

        return '{schema}://{auth}{hostname}{port}/{targets}?{params}'.format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port='' if self.port is None or self.port == default_port
                 else ':{}'.format(self.port),
            targets=','.join(
                [NotifyMQTT.quote(x, safe='/') for x in self.topics]),
            params=NotifyMQTT.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        return len(self.topics)

    @staticmethod
    def parse_url(url):
        """
        There are no parameters nessisary for this protocol; simply having
        windows:// is all you need.  This function just makes sure that
        is in place.

        """

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        try:
            # Acquire topic(s)
            results['targets'] = parse_list(
                NotifyMQTT.unquote(results['fullpath'].lstrip('/')))

        except AttributeError:
            # No 'fullpath' specified
            results['targets'] = []

        # The MQTT protocol version to use
        if 'version' in results['qsd'] and len(results['qsd']['version']):
            results['version'] = \
                NotifyMQTT.unquote(results['qsd']['version'])

        # The MQTT Client ID
        if 'client_id' in results['qsd'] and len(results['qsd']['client_id']):
            results['client_id'] = \
                NotifyMQTT.unquote(results['qsd']['client_id'])

        if 'session' in results['qsd'] and len(results['qsd']['session']):
            results['session'] = parse_bool(results['qsd']['session'])

        # Message Retain Flag
        if 'retain' in results['qsd'] and len(results['qsd']['retain']):
            results['retain'] = parse_bool(results['qsd']['retain'])

        # The MQTT Quality of Service to use
        if 'qos' in results['qsd'] and len(results['qsd']['qos']):
            results['qos'] = \
                NotifyMQTT.unquote(results['qsd']['qos'])

        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'].extend(
                NotifyMQTT.parse_list(results['qsd']['to']))

        # return results
        return results

    @property
    def CA_CERTIFICATE_FILE_LOCATIONS(self):
        """
        Return possible locations to root certificate authority (CA) bundles.

        Taken from https://golang.org/src/crypto/x509/root_linux.go
        TODO: Maybe refactor to a general utility function?
        """
        candidates = [
            # Debian/Ubuntu/Gentoo etc.
            "/etc/ssl/certs/ca-certificates.crt",
            # Fedora/RHEL 6
            "/etc/pki/tls/certs/ca-bundle.crt",
            # OpenSUSE
            "/etc/ssl/ca-bundle.pem",
            # OpenELEC
            "/etc/pki/tls/cacert.pem",
            # CentOS/RHEL 7
            "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem",
            # macOS Homebrew; brew install ca-certificates
            "/usr/local/etc/ca-certificates/cert.pem",
        ]

        # Certifi provides Mozilla’s carefully curated collection of Root
        # Certificates for validating the trustworthiness of SSL certificates
        # while verifying the identity of TLS hosts. It has been extracted from
        # the Requests project.
        try:
            import certifi
            candidates.append(certifi.where())
        except ImportError:  # pragma: no cover
            pass

        return candidates



# -*- coding: utf-8 -*-
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# To use this service you will need a BulkSMS account
# You will need credits (new accounts start with a few)
#     https://www.bulksms.com/account/
#
# API is documented here:
#   - https://www.bulksms.com/developer/json/v1/#tag/Message
import re
import requests
import json
from itertools import chain
from .base import NotifyBase
from ..url import PrivacyMode
from ..common import NotifyType
from ..utils.parse import is_phone_no, parse_phone_no, parse_bool
from ..locale import gettext_lazy as _


IS_GROUP_RE = re.compile(
    r'^(@?(?P<group>[A-Z0-9_-]+))$',
    re.IGNORECASE,
)


class BulkSMSRoutingGroup(object):
    """
    The different categories of routing
    """
    ECONOMY = "ECONOMY"
    STANDARD = "STANDARD"
    PREMIUM = "PREMIUM"


# Used for verification purposes
BULKSMS_ROUTING_GROUPS = (
    BulkSMSRoutingGroup.ECONOMY,
    BulkSMSRoutingGroup.STANDARD,
    BulkSMSRoutingGroup.PREMIUM,
)


class BulkSMSEncoding(object):
    """
    The different categories of routing
    """
    TEXT = "TEXT"
    UNICODE = "UNICODE"
    BINARY = "BINARY"


class NotifyBulkSMS(NotifyBase):
    """
    A wrapper for BulkSMS Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = 'BulkSMS'

    # The services URL
    service_url = 'https://bulksms.com/'

    # All notification requests are secure
    secure_protocol = 'bulksms'

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = 'https://github.com/caronc/apprise/wiki/Notify_bulksms'

    # BulkSMS uses the http protocol with JSON requests
    notify_url = 'https://api.bulksms.com/v1/messages'

    # The maximum length of the body
    body_maxlen = 160

    # The maximum amount of texts that can go out in one batch
    default_batch_size = 4000

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = (
        '{schema}://{user}:{password}@{targets}',
    )

    # Define our template tokens
    template_tokens = dict(NotifyBase.template_tokens, **{
        'user': {
            'name': _('User Name'),
            'type': 'string',
            'required': True,
        },
        'password': {
            'name': _('Password'),
            'type': 'string',
            'private': True,
            'required': True,
        },
        'target_phone': {
            'name': _('Target Phone No'),
            'type': 'string',
            'prefix': '+',
            'regex': (r'^[0-9\s)(+-]+$', 'i'),
            'map_to': 'targets',
        },
        'target_group': {
            'name': _('Target Group'),
            'type': 'string',
            'prefix': '+',
            'regex': (r'^[A-Z0-9 _-]+$', 'i'),
            'map_to': 'targets',
        },
        'targets': {
            'name': _('Targets'),
            'type': 'list:string',
            'required': True,
        },
    })

    # Define our template arguments
    template_args = dict(NotifyBase.template_args, **{
        'to': {
            'alias_of': 'targets',
        },
        'from': {
            'name': _('From Phone No'),
            'type': 'string',
            'regex': (r'^\+?[0-9\s)(+-]+$', 'i'),
            'map_to': 'source',
        },
        'route': {
            'name': _('Route Group'),
            'type': 'choice:string',
            'values': BULKSMS_ROUTING_GROUPS,
            'default': BulkSMSRoutingGroup.STANDARD,
        },
        'unicode': {
            # Unicode characters
            'name': _('Unicode Characters'),
            'type': 'bool',
            'default': True,
        },
        'batch': {
            'name': _('Batch Mode'),
            'type': 'bool',
            'default': False,
        },
    })

    def __init__(self, source=None, targets=None, unicode=None, batch=None,
                 route=None, **kwargs):
        """
        Initialize BulkSMS Object
        """
        super(NotifyBulkSMS, self).__init__(**kwargs)

        self.source = None
        if source:
            result = is_phone_no(source)
            if not result:
                msg = 'The Account (From) Phone # specified ' \
                      '({}) is invalid.'.format(source)
                self.logger.warning(msg)
                raise TypeError(msg)

            # Tidy source
            self.source = '+{}'.format(result['full'])

        # Setup our route
        self.route = self.template_args['route']['default'] \
            if not isinstance(route, str) else route.upper()
        if self.route not in BULKSMS_ROUTING_GROUPS:
            msg = 'The route specified ({}) is invalid.'.format(route)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Define whether or not we should set the unicode flag
        self.unicode = self.template_args['unicode']['default'] \
            if unicode is None else bool(unicode)

        # Define whether or not we should operate in a batch mode
        self.batch = self.template_args['batch']['default'] \
            if batch is None else bool(batch)

        # Parse our targets
        self.targets = list()
        self.groups = list()

        for target in parse_phone_no(targets):
            # Parse each phone number we found
            result = is_phone_no(target)
            if result:
                self.targets.append('+{}'.format(result['full']))
                continue

            group_re = IS_GROUP_RE.match(target)
            if group_re and not target.isdigit():
                # If the target specified is all digits, it MUST have a @
                # in front of it to eliminate any ambiguity
                self.groups.append(group_re.group('group'))
                continue

            self.logger.warning(
                'Dropped invalid phone # and/or Group '
                '({}) specified.'.format(target),
            )

        return

    def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
        """
        Perform BulkSMS Notification
        """

        if not (self.password and self.user):
            self.logger.warning(
                'There were no valid login credentials provided')
            return False

        if not (self.targets or self.groups):
            # We have nothing to notify
            self.logger.warning('There are no BulkSMS targets to notify')
            return False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            'User-Agent': self.app_id,
            'Content-Type': 'application/json',
        }

        # Prepare our payload
        payload = {
            # The To gets populated in the loop below
            'to': None,
            'body': body,
            'routingGroup': self.route,
            'encoding': BulkSMSEncoding.UNICODE
            if self.unicode else BulkSMSEncoding.TEXT,
            # Options are NONE, ALL and ERRORS
            'deliveryReports': "ERRORS"
        }

        if self.source:
            payload.update({
                'from': self.source,
            })

        # Authentication
        auth = (self.user, self.password)

        # Prepare our targets
        targets = list(self.targets) if batch_size == 1 else \
            [self.targets[index:index + batch_size]
             for index in range(0, len(self.targets), batch_size)]
        targets += [{"type": "GROUP", "name": g} for g in self.groups]

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our user
            payload['to'] = target

            # Printable reference
            if isinstance(target, dict):
                p_target = target['name']

            elif isinstance(target, list):
                p_target = '{} targets'.format(len(target))

            else:
                p_target = target

            # Some Debug Logging
            self.logger.debug('BulkSMS POST URL: {} (cert_verify={})'.format(
                self.notify_url, self.verify_certificate))
            self.logger.debug('BulkSMS Payload: {}' .format(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=json.dumps(payload),
                    headers=headers,
                    auth=auth,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                # The responsne might look like:
                # [
                #   {
                #       "id": "string",
                #       "type": "SENT",
                #       "from": "string",
                #       "to": "string",
                #       "body": null,
                #       "encoding": "TEXT",
                #       "protocolId": 0,
                #       "messageClass": 0,
                #       "numberOfParts": 0,
                #       "creditCost": 0,
                #       "submission": {...},
                #       "status": {...},
                #       "relatedSentMessageId": "string",
                #       "userSuppliedId": "string"
                #   }
                # ]

                if r.status_code not in (
                        requests.codes.created, requests.codes.ok):
                    # We had a problem
                    status_str = \
                        NotifyBase.http_response_code_lookup(r.status_code)

                    # set up our status code to use
                    status_code = r.status_code

                    self.logger.warning(
                        'Failed to send BulkSMS notification to {}: '
                        '{}{}error={}.'.format(
                            p_target,
                            status_str,
                            ', ' if status_str else '',
                            status_code))

                    self.logger.debug(
                        'Response Details:\r\n{}'.format(r.content))

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        'Sent BulkSMS notification to {}.'.format(p_target))

            except requests.RequestException as e:
                self.logger.warning(
                    'A Connection error occurred sending BulkSMS: to %s ',
                    p_target)
                self.logger.debug('Socket Exception: %s' % str(e))

                # Mark our failure
                has_error = True
                continue

        return not has_error

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Define any URL parameters
        params = {
            'unicode': 'yes' if self.unicode else 'no',
            'batch': 'yes' if self.batch else 'no',
            'route': self.route,
        }

        if self.source:
            params['from'] = self.source

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return '{schema}://{user}:{password}@{targets}/?{params}'.format(
            schema=self.secure_protocol,
            user=self.pprint(self.user, privacy, safe=''),
            password=self.pprint(
                self.password, privacy, mode=PrivacyMode.Secret, safe=''),
            targets='/'.join(chain(
                [NotifyBulkSMS.quote('{}'.format(x), safe='+')
                 for x in self.targets],
                [NotifyBulkSMS.quote('@{}'.format(x), safe='@')
                 for x in self.groups])),
            params=NotifyBulkSMS.urlencode(params))

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (
            self.secure_protocol,
            self.user if self.user else None,
            self.password if self.password else None,
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """

        #
        # Factor batch into calculation
        #
        # Note: Groups always require a separate request (and can not be
        # included in batch calculations)
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + \
                (1 if targets % batch_size else 0)

        return targets + len(self.groups)

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results['targets'] = [
            NotifyBulkSMS.unquote(results['host']),
            *NotifyBulkSMS.split_path(results['fullpath'])]

        # Support the 'from'  and 'source' variable so that we can support
        # targets this way too.
        # The 'from' makes it easier to use yaml configuration
        if 'from' in results['qsd'] and len(results['qsd']['from']):
            results['source'] = \
                NotifyBulkSMS.unquote(results['qsd']['from'])

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if 'to' in results['qsd'] and len(results['qsd']['to']):
            results['targets'] += \
                NotifyBulkSMS.parse_phone_no(results['qsd']['to'])

        # Unicode Characters
        results['unicode'] = \
            parse_bool(results['qsd'].get(
                'unicode', NotifyBulkSMS.template_args['unicode']['default']))

        # Get Batch Mode Flag
        results['batch'] = \
            parse_bool(results['qsd'].get(
                'batch', NotifyBulkSMS.template_args['batch']['default']))

        # Allow one to define a route group
        if 'route' in results['qsd'] and len(results['qsd']['route']):
            results['route'] = \
                NotifyBulkSMS.unquote(results['qsd']['route'])

        return results
